# --------------------------------------------------------------------
# shell.py
#
# Author: Lain Musgrove (lain.musgrove@hearst.com)
# Date: Sunday August 27, 2023
# --------------------------------------------------------------------
import shlex
from enum import Enum
from pathlib import Path
from typing import Iterable, Optional, cast

from xeno.recipe import Events, Recipe, recipe
from xeno.shell import Environment, PathSpec, Shell
from xeno.utils import is_iterable


# --------------------------------------------------------------------
class ShellRecipe(Recipe):
    class Result(Enum):
        CODE = 0
        STDOUT = 1
        STDERR = 2
        FILE = 3

    ResultSpec = Iterable[Result] | Result

    class Format(Recipe.Format):
        def sigil(self, recipe: Recipe) -> str:
            assert isinstance(recipe, ShellRecipe)
            recipe = cast(ShellRecipe, recipe)
            if recipe.has_target():
                return f"{recipe.program_name()}:{recipe.rel_target()}"
            else:
                return f"{recipe.program_name()}"

        def start(self, recipe: Recipe) -> str:
            assert isinstance(recipe, ShellRecipe)
            recipe = cast(ShellRecipe, recipe)
            cmd = recipe.shell.interpolate(
                recipe.cmd, recipe.scanner.kwargs(Recipe.PassMode.TARGETS)
            )
            return cmd

    def __init__(
        self,
        cmd,
        *,
        as_user: Optional[str] = None,
        cleanup: Optional[str | list[str]] = None,
        cleanup_cwd: Optional[PathSpec] = None,
        code=0,
        ctrlc=False,
        cwd: Optional[PathSpec | Recipe] = None,
        env: Optional[Environment] = None,
        interact=False,
        memoize=False,
        name: Optional[str] = None,
        quiet=False,
        result: Optional["ShellRecipe.ResultSpec"] = None,
        redacted: set[str] = set(),
        sync=False,
        **kwargs,
    ):
        shell_cwd = cwd
        if isinstance(cwd, Recipe):
            shell_cwd = cwd.target

        self.env = Environment.context() if env is None else {**env}
        self.shell = Shell(self.env, shell_cwd)
        self.cmd: str | list[str] = []
        self.cleanup_cmd = cleanup
        self.cleanup_cwd = Path(cleanup_cwd or Path.cwd())

        if cwd:
            # Add 'cwd' to kwargs if specified, as kwargs is used
            # by the scanner to interpolate keyword args into
            # commands and determine dependencies.
            kwargs["cwd"] = cwd

        target = None
        component_list = []

        if Recipe.DEFAULT_TARGET_PARAM in kwargs:
            target = kwargs[Recipe.DEFAULT_TARGET_PARAM]

        def convert_cmd(cmd_comp) -> str:
            if isinstance(cmd_comp, str | Path):
                return str(cmd_comp)

            elif isinstance(cmd_comp, Recipe):
                assert (
                    cmd_comp.has_target()
                ), "Recipe without a target can't be used in a shell command."
                component_list.append(cmd_comp)
                target = cmd_comp.rel_target()
                if Path.cwd() == target.parent.absolute():
                    return f"./{target}"
                return str(target)

            else:
                raise ValueError(f"Invalid shell command component: {cmd_comp}")

        if is_iterable(cmd):
            self.cmd = [convert_cmd(c) for c in cmd]
        else:
            self.cmd = convert_cmd(cmd)

        self.expected_code = code
        self.interact = interact
        self.ctrlc = ctrlc
        self.quiet = quiet
        self.redacted = redacted
        self.result_spec = result
        self.scanner = Recipe.scan([], kwargs)
        super().__init__(
            component_list=component_list,
            component_map=self.scanner.component_map(),
            as_user=as_user,
            memoize=memoize,
            name=name,
            fmt=ShellRecipe.Format(),
            static_files=self.scanner.paths(),
            sync=sync,
            target=target,
        )

        self.name = self.name or self.program_name()

        self.return_code: Optional[int] = None
        self.stdout_lines: list[str] = []
        self.stderr_lines: list[str] = []

    def program_name(self):
        if isinstance(self.cmd, list):
            cmd = self.cmd[0]
        else:
            cmd = shlex.split(self.cmd)[0]
        try:
            cmd = str(Path(cmd).relative_to(Path.cwd()))
        except ValueError:
            pass
        return self.shell.interpolate(cmd, {})

    def log_stdout(self, line: str, _):
        if not self.quiet:
            self.log(Events.INFO, line)
        self.stdout_lines.append(line)

    def log_stderr(self, line: str, _):
        if not self.quiet:
            self.log(Events.WARNING, line)
        self.stderr_lines.append(line)

    async def clean(self):
        await super().clean()

        if self.cleanup_cmd:
            self._make_interactive(True)
            if self.return_code != 0:
                raise RuntimeError("Failed to run cleanup command.")

    async def make(self):
        self.return_code = None
        self.stdout_lines.clear()
        self.stderr_lines.clear()

        if self.interact or self.as_user:
            # Do interactive mode so we can `sudo`.
            self._make_interactive()

        else:
            self.return_code = await self.shell.run(
                self.cmd,
                stdout=self.log_stdout,
                stderr=self.log_stderr,
                **self.scanner.kwargs(Recipe.PassMode.RESULTS),
            )

        assert (
            self.return_code == self.expected_code
        ), f"Unexpected return code.  (expected {self.expected_code}, got {self.return_code})"

        return self._compute_result()

    def _make_interactive(self, cleanup=False):
        shell = self.shell

        if cleanup:
            cmd = self.cleanup_cmd
            shell = shell.cd(self.cleanup_cwd)
            pass_mode = Recipe.PassMode.TARGETS
        else:
            cmd = self.cmd
            pass_mode = Recipe.PassMode.RESULTS

        assert cmd

        kwargs = self.scanner.kwargs(pass_mode)
        if self.as_user:
            self.return_code = shell.interact_as(
                self.as_user, cmd, ctrlc=self.ctrlc, **kwargs
            )
        else:
            self.return_code = shell.interact(cmd, ctrlc=self.ctrlc, **kwargs)

    def _get_result(self, spec: Result):
        match spec:
            case ShellRecipe.Result.CODE:
                return self.return_code
            case ShellRecipe.Result.FILE:
                return self.target
            case ShellRecipe.Result.STDOUT:
                result = self.stdout_lines
                self.stdout_lines = []
                return result
            case ShellRecipe.Result.STDERR:
                result = self.stderr_lines
                self.stderr_lines = []
                return result

    def _compute_result(self):
        if self.result_spec is None:
            return self._get_result(
                ShellRecipe.Result.FILE
                if self.has_target()
                else ShellRecipe.Result.CODE
            )

        if isinstance(self.result_spec, ShellRecipe.Result):
            return self._get_result(self.result_spec)

        return [self._get_result(r) for r in self.result_spec]


# --------------------------------------------------------------------
@recipe(factory=True)
def sh(
    cmd,
    **kwargs,
):
    if "env" in kwargs:
        kwargs["env"] = {**sh.env, **kwargs["env"]}
    elif sh.env:
        kwargs["env"] = sh.env
    return ShellRecipe(cmd, **kwargs)


# --------------------------------------------------------------------
sh.env = {}
sh.result = ShellRecipe.Result



# --------------------------------------------------------------------
# errors.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------


# --------------------------------------------------------------------
class InjectionError(Exception):
    pass


# --------------------------------------------------------------------
class MissingResourceError(InjectionError):
    def __init__(self, name):
        super().__init__('The resource "%s" was not provided.' % name)
        self.name = name


# --------------------------------------------------------------------
class MissingDependencyError(InjectionError):
    def __init__(self, name, dep_name):
        super().__init__(
            f'Resource "{dep_name}" required by "{name}" was not provided.')
        self.name = name
        self.dep_name = dep_name


# --------------------------------------------------------------------
class MethodInjectionError(InjectionError):
    def __init__(self, method, name, reason=None):
        super().__init__(
            f'Failed to inject "{name}" into {method.__qualname__}'
            + f': {reason}' if reason else '.'
        )
        self.method = method
        self.name = name


# --------------------------------------------------------------------
class ClassInjectionError(InjectionError):
    def __init__(self, class_, name, reason=None):
        super().__init__(
            'Failed to inject "%s" into constructor for class "%s".'
            % (name, class_.__qualname__)
            + reason
            or ""
        )
        self.class_ = class_
        self.name = name


# --------------------------------------------------------------------
class CircularDependencyError(InjectionError):
    def __init__(self, resource, dep):
        super().__init__(
            'Circular dependency detected between "%s" and "%s".' % (
                resource, dep)
        )
        self.resource = resource
        self.dep = dep


# --------------------------------------------------------------------
class UndefinedNameError(InjectionError):
    def __init__(self, name):
        super().__init__('Undefined name: "%s"' % name)
        self.name = name


# --------------------------------------------------------------------
class UnknownNamespaceError(InjectionError):
    def __init__(self, name):
        super().__init__('Unknown namespace: "%s"' % name)
        self.name = name


# --------------------------------------------------------------------
class InvalidResourceError(InjectionError):
    pass



# --------------------------------------------------------------------
# async_injector.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import inspect
from collections import defaultdict
from typing import Optional, cast

from .abstract import AbstractInjector
from .attributes import (
    NOTHING,
    ClassAttributes,
    MethodAttributes,
    get_injection_params,
    get_injection_points,
)
from .decorators import Tags, named, singleton
from .errors import (
    ClassInjectionError,
    MethodInjectionError,
    MissingDependencyError,
    MissingResourceError,
)
from .namespaces import Namespace
from .utils import async_map, async_wrap, resolve_alias


# --------------------------------------------------------------------
class AsyncInjector(AbstractInjector):
    """
    A specialization of AbstractInjector, and the classic Injector
    which supports async providers and uses an asyncio event loop
    during dependency resolution.

    Use of this injector is preferred for flexibility if you know
    that it will not be run within an asyncio event loop.

    This form of Injector cannot be used within other event loops.
    If you wish to perform synchronous dependency injection inside
    of an asyncio event loop, use SyncInjector instead.
    """

    def __init__(self, *modules):
        """
        Create an Injector object.

        *modules: A list of modules to include in the injector.
                  More modules can be added later by calling
                  Injector.add_module().
        """
        self.async_injection_interceptors = []
        self.singleton_locks = defaultdict(asyncio.Lock)
        super().__init__(*modules)

    def add_async_injection_interceptor(self, interceptor):
        self.async_injection_interceptors.append(interceptor)

    def create(self, class_):
        """
        Overrides: AbstractInjector

        This method creates a new event loop.  If running from an existing
        event loop, await `create_async` instead.

        Create an instance of the specified class.  The class' constructor
        must follow the rules for @inject methods, such that all of its
        parameters refer to injectable resources or are optional.

        If the object needs to be constructed with objects not from the
        Injector, do not use this method.  Instead, instantiate the object
        with these parameters in the constructor, then mark one or more
        methods with @inject and pass the instance to Injector.inject().
        """
        return asyncio.run(self.create_async(class_))

    async def create_async(self, class_):
        """
        Create an instance of the specified class.  The class' constructor
        must follow the rules for @inject methods, such that all of its
        parameters refer to injectable resources or are optional.

        If the object needs to be constructed with objects not from the
        Injector, do not use this method.  Instead, instantiate the object
        with these parameters in the constructor, then mark one or more
        methods with @inject and pass the instance to Injector.inject().
        """
        try:
            param_map, alias_map = await self._resolve_dependencies(
                class_.__init__, unbound_ctor=True
            )
            attrs = MethodAttributes.for_method(class_.__init__)
            param_map = await self._async_invoke_injection_interceptors(
                attrs, param_map, alias_map
            )
        except MethodInjectionError as e:
            raise ClassInjectionError(class_, e.name)

        instance = class_(**param_map)
        await self._inject_instance(instance)
        return instance

    def inject(self, obj, aliases={}, namespace=""):
        """
        Overrides: AbstractInjector

        This method creates a new event loop.  If running from an existing
        event loop, await `inject_async` instead.

        Inject a method or object instance with resources from this Injector.

        obj: A method or object instance.  If this is a method, all named
             parameters are injected from the Injector.  If this is an
             instance, its methods are scanned for injection points and these
             methods are all invoked with resources from the Injector.
        aliases: An optional map from dependency alias to real dependency name.
        """
        return asyncio.run(self.inject_async(obj, aliases, namespace))

    async def inject_async(self, obj, aliases={}, namespace: Optional[str] = None):
        """
        Inject a method or object instance with resources from this Injector.

        obj: A method or object instance.  If this is a method, all named
             parameters are injected from the Injector.  If this is an
             instance, its methods are scanned for injection points and these
             methods are all invoked with resources from the Injector.
        aliases: An optional map from dependency alias to real dependency name.
        """
        if inspect.isfunction(obj) or inspect.ismethod(obj):
            return await self._inject_method(obj, aliases, namespace or "")
        return await self._inject_instance(obj, aliases, namespace or "")

    def require(self, name, method=None):
        """
        Overrides: AbstractInjector

        This method creates a new event loop.  If running from an existing
        event loop, await `require_async` instead.

        Require a named resource from this Injector.  If it can't be provided,
        an InjectionError is raised.

        Optionally, a method can be specified.  This indicates the method that
        requires the resource for injection, and causes this method to throw
        MethodInjectionError instead of InjectionError if the resource was
        not provided.
        """
        return asyncio.run(self.require_async(name, method))

    async def require_async(self, name, method=None):
        """
        Require a named resource from this Injector.  If it can't be provided,
        an InjectionError is raised.

        Optionally, a method can be specified.  This indicates the method that
        requires the resource for injection, and causes this method to throw
        MethodInjectionError instead of InjectionError if the resource was
        not provided.
        """
        return await self._require_coro(name, method)

    def provide(
        self, name_or_method, value=NOTHING, is_singleton=False, namespace=None
    ):
        """
        Overrides: AbstractInjector

        This method creates a new event loop.  If running from an existing
        event loop, await `provide_async` instead.
        """

        return asyncio.run(
            self.provide_async(name_or_method, value, is_singleton, namespace)
        )

    async def provide_async(
        self, name_or_method, value=NOTHING, is_singleton=False, namespace=None
    ):

        if inspect.ismethod(name_or_method) or inspect.isfunction(name_or_method):
            value = name_or_method
            attrs = MethodAttributes.for_method(name_or_method)
            assert attrs is not None
            name = attrs.get(Tags.NAME)
        elif value is NOTHING:
            raise ValueError("A name and value or just a method must be provided.")
        else:
            name = name_or_method

        if name in self.singletons:
            del self.singletons[name]
        if inspect.ismethod(value) or inspect.isfunction(value):
            if is_singleton:
                value = singleton(value)
            await self._bind_resource_async(value, namespace=namespace)
        else:

            @named(name)
            def wrapper():
                return value

            if is_singleton:
                wrapper = singleton(wrapper)
            await self._bind_resource_async(wrapper, namespace=namespace)

    async def _bind_resource_async(
        self, bound_method, module_aliases={}, namespace=None
    ):
        params, _ = get_injection_params(bound_method)
        attrs = MethodAttributes.for_method(bound_method)
        assert attrs is not None

        using_namespaces = []
        name = cast(str, attrs.get(Tags.NAME))
        assert name is not None
        # Allow names that begin with the namespace separator
        # to be scoped outside of the specified namespace.
        if name.startswith(Namespace.SEP):
            name = name[len(Namespace.SEP):]
        elif namespace is not None:
            name = Namespace.join(namespace, name)
            using_namespaces.append(namespace)

        aliases = {
            **(self._get_aliases(attrs, using_namespaces) or {}),
            **module_aliases,
        }

        injected_method = await self.inject_async(bound_method, aliases, namespace)

        if attrs.check(Tags.SINGLETON):
            async def wrapper():
                async with self.singleton_locks[name]:
                    if name not in self.singletons:
                        singleton = await injected_method()
                        self.singletons[name] = singleton
                        return singleton
                    return self.singletons[name]

            resource = wrapper
        else:
            resource = injected_method

        attrs.put(Tags.RESOURCE_FULL_NAME, name)

        self.ns_index.add(name)
        self.resources[name] = resource
        self.resource_attrs[name] = attrs
        self.dep_graph[name] = lambda: [
            resolve_alias(x, aliases) for x in params
        ]

    def _bind_resource(self, bound_method, module_aliases={}, namespace=None):
        """
        Overrides: AbstractInjector
        """
        return asyncio.run(
            self._bind_resource_async(bound_method, module_aliases, namespace)
        )

    async def _resolve_dependencies(
        self, f, unbound_ctor=False, aliases={}, namespace=""
    ):
        params, default_set = get_injection_params(f, unbound_ctor=unbound_ctor)
        attrs = MethodAttributes.for_method(f)
        assert attrs is not None
        param_map: dict = {}
        param_resource_map = {}
        full_name = attrs.get(Tags.NAME)
        if namespace:
            full_name = Namespace.join(namespace, full_name)
            aliases = {**aliases, **self._get_aliases(attrs, [namespace])}

        try:
            resource_async_map = {}
            for param in params:
                if param in default_set and not self.has(param):
                    continue
                resource_name = param
                if resource_name.startswith(Namespace.SEP):
                    resource_name = resource_name[len(Namespace.SEP):]
                resource_name = resolve_alias(resource_name, aliases)
                resource_async_map[param] = self._require_coro(resource_name)
                param_resource_map[param] = resource_name

            param_map = dict(
                await asyncio.gather(
                    *(async_map(k, c) for k, c in resource_async_map.items())
                )
            )

        except MissingResourceError as e:
            raise MissingDependencyError(full_name, e.name) from e
        return param_map, param_resource_map

    async def _inject_instance(self, instance, aliases={}, namespace=""):
        class_attributes = ClassAttributes.for_class(instance.__class__)
        assert class_attributes is not None
        aliases = {**aliases, **class_attributes.get(Tags.ALIASES, {})}
        for _, injection_point in get_injection_points(instance):
            injected_method = await self.inject_async(
                injection_point, aliases, namespace
            )
            await injected_method()
        return instance

    async def _inject_method(self, method, aliases_in={}, namespace=""):
        async def wrapper():
            aliases = {**aliases_in}
            attrs = MethodAttributes.for_method(method)
            assert attrs is not None
            aliases = {**aliases, **attrs.get(Tags.ALIASES, {})}
            param_map, alias_map = await self._resolve_dependencies(
                method, aliases=aliases, namespace=namespace
            )
            param_map = await self._async_invoke_injection_interceptors(
                attrs, param_map, alias_map
            )
            return await async_wrap(method, **param_map)

        return wrapper

    async def _async_invoke_injection_interceptors(self, attrs, param_map, alias_map):
        param_map = self._invoke_injection_interceptors(attrs, param_map, alias_map)
        for interceptor in self.async_injection_interceptors:
            param_map = await interceptor(attrs, param_map, alias_map)
        return param_map

    async def _require_coro(self, name, method=None):
        """
        For internal use only.  Used to tie together resources needed
        by other resources in this injector.
        """
        if name not in self.resources:
            if method is not None:
                raise MethodInjectionError(method, name, "Resource was not provided.")
            raise MissingResourceError(name)
        if name in self.singletons:
            return self.singletons[name]
        return await async_wrap(self.resources[name])



# --------------------------------------------------------------------
# sync.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import inspect
from typing import Optional, cast

from .abstract import AbstractInjector
from .attributes import (NOTHING, ClassAttributes, MethodAttributes,
                         get_injection_params, get_injection_points)
from .decorators import Tags, named, singleton
from .errors import (ClassInjectionError, MethodInjectionError,
                     MissingDependencyError, MissingResourceError)
from .namespaces import Namespace
from .utils import resolve_alias


# --------------------------------------------------------------------
class SyncInjector(AbstractInjector):
    """
    SyncInjector is a specialization of AbstractInjector that does not support
    async providers and can be run inside another event loop.
    """

    def create(self, class_):
        """
        Overrides: AbstractInjector
        """
        try:
            param_map, alias_map = self._resolve_dependencies(
                class_.__init__, unbound_ctor=True
            )
            attrs = MethodAttributes.for_method(class_.__init__)
            param_map = self._invoke_injection_interceptors(attrs, param_map, alias_map)
        except MethodInjectionError as e:
            raise ClassInjectionError(class_, e.name)

        instance = class_(**param_map)
        self._inject_instance(instance)
        return instance

    def inject(self, obj, aliases={}, namespace: Optional[str] = None):
        """
        Overrides: AbstractInjector
        """
        if inspect.isfunction(obj) or inspect.ismethod(obj):
            return self._inject_method(obj, aliases, namespace or "")
        return self._inject_instance(obj, aliases, namespace or "")

    def require(self, name, method=None):
        """
        Overrides: AbstractInjector
        """
        return self._require(name, method)

    def provide(self, name_or_method, value=NOTHING, is_singleton=False, namespace=None):
        """
        Overrides: AbstractInjector
        """

        if inspect.ismethod(name_or_method) or inspect.isfunction(name_or_method):
            value = name_or_method
            attrs = MethodAttributes.for_method(name_or_method)
            assert attrs is not None
            name = attrs.get(Tags.NAME)
        elif value is NOTHING:
            raise ValueError("A name and value or just a method must be provided.")
        else:
            name = name_or_method

        if name in self.singletons:
            del self.singletons[name]
        if inspect.ismethod(value) or inspect.isfunction(value):
            if is_singleton:
                value = singleton(value)
            self._bind_resource(value, namespace=namespace)
        else:
            @named(name)
            def wrapper():
                return value

            if is_singleton:
                wrapper = singleton(wrapper)
            self._bind_resource(wrapper, namespace=namespace)

    def _bind_resource(self, bound_method, module_aliases={}, namespace=None):
        """
        Overrides: AbstractInjector
        """

        params, _ = get_injection_params(bound_method)
        attrs = MethodAttributes.for_method(bound_method)
        assert attrs is not None

        using_namespaces = []
        name = cast(str, attrs.get(Tags.NAME))
        # Allow names that begin with the namespace separator
        # to be scoped outside of the specified namespace.
        if name.startswith(Namespace.SEP):
            name = name[len(Namespace.SEP):]
        elif namespace is not None:
            name = Namespace.join(namespace, name)
            using_namespaces.append(namespace)

        aliases = {
            **(self._get_aliases(attrs, using_namespaces) or {}),
            **module_aliases,
        }

        injected_method = self.inject(bound_method, aliases, namespace)

        if attrs.check(Tags.SINGLETON):
            def wrapper():
                if name not in self.singletons:
                    singleton = injected_method()
                    self.singletons[name] = singleton
                    return singleton
                return self.singletons[name]

            resource = wrapper
        else:
            resource = injected_method

        # Make the canonical full resource name available via 'resource-name'.
        attrs.put(Tags.RESOURCE_FULL_NAME, name)

        self.ns_index.add(name)
        self.resources[name] = resource
        self.resource_attrs[name] = attrs
        self.dep_graph[name] = lambda: [
            resolve_alias(x, aliases) for x in params
        ]

    def _resolve_dependencies(self, f, unbound_ctor=False, aliases={}, namespace=""):
        params, default_set = get_injection_params(f, unbound_ctor=unbound_ctor)
        attrs = MethodAttributes.for_method(f)
        assert attrs is not None
        param_map = {}
        param_resource_map = {}
        full_name = attrs.get(Tags.NAME)
        if namespace:
            full_name = Namespace.join(namespace, full_name)
            aliases = {**aliases, **self._get_aliases(attrs, [namespace])}

        try:
            resource_map = {}
            for param in params:
                if param in default_set and not self.has(param):
                    continue
                resource_name = param
                if resource_name.startswith(Namespace.SEP):
                    resource_name = resource_name[len(Namespace.SEP):]
                resource_name = resolve_alias(resource_name, aliases)
                resource_map[param] = self._require(resource_name)
                param_resource_map[param] = resource_name

            for k, c in resource_map.items():
                param_map[k] = c

        except MissingResourceError as e:
            raise MissingDependencyError(full_name, e.name) from e
        return param_map, param_resource_map

    def _require(self, name, method=None):
        if name not in self.resources:
            if method is not None:
                raise MethodInjectionError(method, name, "Resource was not provided.")
            raise MissingResourceError(name)
        if name in self.singletons:
            return self.singletons[name]
        return self.resources[name]()

    def _inject_instance(self, instance, aliases={}, namespace=""):
        class_attributes = ClassAttributes.for_class(instance.__class__)
        assert class_attributes is not None
        aliases = {**aliases, **class_attributes.get(Tags.ALIASES, {})}
        for _, injection_point in get_injection_points(instance):
            injected_method = self.inject(injection_point, aliases, namespace)
            injected_method()
        return instance

    def _inject_method(self, method, aliases_in={}, namespace=""):
        def wrapper():
            aliases = {**aliases_in}
            attrs = MethodAttributes.for_method(method)
            assert attrs is not None
            aliases = {**aliases, **attrs.get(Tags.ALIASES, {})}
            param_map, alias_map = self._resolve_dependencies(
                method, aliases=aliases, namespace=namespace
            )
            param_map = self._invoke_injection_interceptors(
                attrs, param_map, alias_map
            )
            return method(**param_map)
        return wrapper



# --------------------------------------------------------------------
# pkg_config.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Saturday March 4, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

from xeno.shell import Environment, check


# --------------------------------------------------------------------
def pkg_config(names: list[str], static=False) -> Environment:
    env = Environment()
    for name in names:
        env.append(PackageConfig(name))
    return env


# --------------------------------------------------------------------
class PackageConfig(Environment):
    """
    An environment wrapper providing build information sourced from pkgconf(1).
    """

    def __init__(self, name: str, static=False):
        self.name = name
        self.static = static
        check(["pkgconf", "--exists", self.name])
        self.cflags = self._get_cflags()
        self.ldflags = self._get_ldflags()
        self.version = self._get_version()
        self["LDFLAGS"] = self.ldflags
        self["CFLAGS"] = self.cflags

    def _get_cflags(self):
        argv = ["pkgconf", "--cflags", self.name]
        if self.static:
            argv.append("--static")

        return check(argv)

    def _get_ldflags(self):
        argv = ["pkgconf", "--libs", self.name]
        if self.static:
            argv.append("--static")

        return check(argv)

    def _get_version(self):
        return check(["pkgconf", "--modversion", self.name]).strip()



# --------------------------------------------------------------------
# c.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Monday August 28, 2023
# --------------------------------------------------------------------

"""
Batteries-included tools and opinionated defaults for building C programs.
"""

from pathlib import Path
from typing import Iterable, Optional

from xeno.recipe import recipe
from xeno.recipes.shell import sh
from xeno.shell import select_env
from xeno.typedefs import NestedIterable, PathSpec
from xeno.utils import expand

# --------------------------------------------------------------------
ENV = select_env("PATH", "LDFLAGS", CC="clang").append(CFLAGS=("-Wall", "--std=c17"))


# -------------------------------------------------------------------
@recipe(factory=True, sigil=lambda r: f"{r.name}:{r.target.name}")
def compile(
    *sources: PathSpec | NestedIterable[PathSpec],
    obj=False,
    headers: Iterable[PathSpec] = [],
    target: Optional[PathSpec] = None,
    env=ENV,
):
    src, *srcs = expand(*sources)

    if obj:
        cmd = "{CC} {CFLAGS} -c {src} {srcs} {LDFLAGS} -o {target}"
        suffix = ".o"
    else:
        cmd = "{CC} {CFLAGS} {src} {srcs} {LDFLAGS} -o {target}"
        suffix = ""

    if target is None:
        target = src

    return sh(cmd, env=env, src=src, srcs=srcs, target=Path(target).with_suffix(suffix))



# --------------------------------------------------------------------
# cxx.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Tuesday August 29, 2023
# --------------------------------------------------------------------

from typing import Iterable, Optional

from xeno.recipes.c import compile as base_compile
from xeno.shell import select_env
from xeno.typedefs import PathSpec

# --------------------------------------------------------------------
ENV = select_env("PATH", "LDFLAGS", CC="clang++").append(
    CFLAGS=("-Wall", "--std=c++2a")
)


# --------------------------------------------------------------------
def compile(
    *sources: PathSpec,
    obj=False,
    headers: Iterable[PathSpec] = [],
    target: Optional[PathSpec] = None,
    env=ENV,
):
    return base_compile(*sources, obj=obj, headers=headers, target=target, env=env)



# --------------------------------------------------------------------
# recipe.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday March 9, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import inspect
import os
import sys
import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Generator, Iterable, Optional, cast, no_type_check

from xeno.attributes import MethodAttributes
from xeno.events import Event, EventBus
from xeno.shell import PathSpec, remove_paths
from xeno.utils import async_map, async_vwrap, async_wrap, is_iterable, list_or_delim

# --------------------------------------------------------------------
UNICODE_SUPPORT = sys.stdout.encoding.lower().startswith("utf")


# --------------------------------------------------------------------
class Events:
    CLEAN = "clean"
    DEBUG = "debug"
    ERROR = "error"
    FAIL = "fail"
    INFO = "info"
    START = "start"
    SUCCESS = "success"
    WARNING = "warning"
    END = "end"


# --------------------------------------------------------------------
class BuildError(Exception):
    def __init__(self, recipe: "Recipe", msg=""):
        super().__init__(f"[{recipe.sigil()}] {msg}")
        self.recipe = recipe


# --------------------------------------------------------------------
class CompositeError(Exception):
    def __init__(
        self, exceptions: Iterable[Exception], msg: str = "Multiple errors occurred."
    ):
        self.msg = msg
        self.exceptions = list(exceptions)
        super().__init__(self._compose_message())

    def _indent_line(self, line: str, indent: int):
        return ("    " * indent) + line

    def _compose_message(self, indent=0):
        sb = []

        sb.append(self.msg)
        for exc in self.exceptions:
            header = f"{exc.__class__.__qualname__}: "
            lines = str(exc).split("\n")
            lines[0] = header + lines[0]

            for line in lines:
                sb.append("    " + line)

        return "\n".join(sb)


# --------------------------------------------------------------------
ListedComponents = list["Recipe"]
ComponentIterable = Iterable["Recipe"]
MappedComponents = dict[str, "Recipe" | Iterable["Recipe"]]
FormatF = Callable[["Recipe"], str]


# --------------------------------------------------------------------
class Recipe:
    DEFAULT_TARGET_PARAM = "target"
    SIGIL_DELIMITER = ":"

    active: set["Recipe"] = set()

    class Format:
        @dataclass
        class Symbols:
            target: str
            start: str
            ok: str
            fail: str
            warning: str

        def __init__(self, unicode_symbols=True):
            if UNICODE_SUPPORT and unicode_symbols:
                self.symbols = self.Symbols(
                    target="->", start="⚙", ok="✓", fail="✗", warning="⚠"
                )
            else:
                self.symbols = self.Symbols(
                    target="->", start="(*)", ok=":)", fail=":(", warning="/!\\"
                )

        def clean(self, recipe: "Recipe") -> str:
            sb: list[str | Path] = []
            sb.append("cleaned")
            if recipe.has_target():
                sb.append(recipe.target)
            return " ".join([str(s) for s in sb])

        def fail(self, recipe: "Recipe") -> str:
            return f"{self.symbols.fail} fail"

        def ok(self, recipe: "Recipe") -> str:
            return f"{self.symbols.ok} ok"

        def sigil(self, recipe: "Recipe") -> str:
            if recipe.has_target():
                return f"{recipe.name}{Recipe.SIGIL_DELIMITER}{recipe.rel_target()}"
            return recipe.name

        def start(self, recipe: "Recipe") -> str:
            return f"{self.symbols.start} start"

    class FormatOverride(Format):
        def __init__(self, fmt: "Recipe.Format", **overrides):
            self.fmt = fmt
            for key in overrides.keys():
                if not hasattr(fmt, key):
                    raise ValueError(key)
            self.overrides = overrides

        def add_override(self, key: str, value: Any):
            self.overrides[key] = value

        def __getattribute__(self, name):
            if name in ("overrides", "fmt", "add_override"):
                return super().__getattribute__(name)
            if name in self.overrides:
                return self.overrides[name]
            else:
                return self.fmt.__getattribute__(name)

    class PassMode(Enum):
        NORMAL = 0
        RESULTS = 1
        TARGETS = 2

    class ParamType(Enum):
        NORMAL = 0
        RECIPE = 1
        PATH = 2

    class Scanner:
        def __init__(self):
            self._args: list[Any] = []
            self._kwargs: dict[str, Any] = {}
            self._arg_offsets: list[int] = []
            self._kwarg_keys: list[str] = []
            self._paths: list[Path] = []

        def clear(self):
            self._args.clear()
            self._kwargs.clear()
            self._arg_offsets.clear()
            self._kwarg_keys.clear()
            self._paths.clear()

        def _scan_one(self, arg) -> tuple[Any, "Recipe.ParamType"]:
            if isinstance(arg, Recipe):
                return arg, Recipe.ParamType.RECIPE

            if is_iterable(arg):
                arg = [*arg]
                if arg and all(isinstance(x, Recipe) for x in arg):
                    return arg, Recipe.ParamType.RECIPE
                if arg and all(isinstance(x, Path) for x in arg):
                    return arg, Recipe.ParamType.PATH

            if isinstance(arg, Path):
                return arg, Recipe.ParamType.PATH

            return arg, Recipe.ParamType.NORMAL

        def has_recipes(self):
            return self._arg_offsets or self._kwarg_keys

        def num_recipes(self):
            return len(self._arg_offsets) + len(self._kwarg_keys)

        async def gather_args(self):
            """
            Await resolution of recipes in args and update args with their results.
            """
            results = await asyncio.gather(
                *[v() if isinstance(v, Recipe) else async_vwrap(v) for v in self._args]
            )
            self.scan_params(*results)

        def bind(self, f: Callable, mode: "Recipe.PassMode"):
            """
            Bind args and kwargs to the given callable.
            """
            return inspect.signature(f).bind(*self.args(mode), **self.kwargs(mode))

        async def gather_kwargs(self):
            """
            Await resolution of recipes in kwargs and update with their results.
            """
            result_tuples = asyncio.gather(
                *[
                    async_map(k, v() if isinstance(v, Recipe) else async_vwrap(v))
                    for k, v in self._kwargs.items()
                ]
            )
            results = {k: v for k, v in result_tuples}
            self.scan_params(**results)

        def gather_all(self):
            """
            Await resolution of recipes in args and kwargs and update both with
            their results.
            """

            return asyncio.gather(self.gather_args(), self.gather_kwargs())

        def scan_params(self, *args, **kwargs):
            self.clear()

            for i, arg in enumerate(args):
                self.scan(arg, offset=i)

            for k, arg in kwargs.items():
                self.scan(arg, key=k)

        def scan(
            self, arg, offset: Optional[int] = None, key: Optional[str] = None
        ) -> "Recipe.ParamType":
            arg, param_type = self._scan_one(arg)

            if param_type == Recipe.ParamType.PATH:
                if is_iterable(arg):
                    self._paths.extend(arg)
                else:
                    self._paths.append(arg)

            if offset is not None:
                self._args.append(arg)
                if param_type == Recipe.ParamType.RECIPE:
                    self._arg_offsets.append(offset)

            elif key is not None:
                self._kwargs[key] = arg
                if param_type == Recipe.ParamType.RECIPE:
                    self._kwarg_keys.append(key)

            return param_type

        def args(self, pass_mode: "Recipe.PassMode") -> list[Any]:
            match pass_mode:
                case Recipe.PassMode.NORMAL:
                    return self._args
                case Recipe.PassMode.RESULTS:
                    results = [*self._args]
                    for offset in self._arg_offsets:
                        value = results[offset]
                        if is_iterable(value):
                            results[offset] = [r.result() for r in value]
                        else:
                            results[offset] = value.result()
                    return results
                case Recipe.PassMode.TARGETS:
                    results = [*self._args]
                    for offset in self._arg_offsets:
                        value = results[offset]
                        if is_iterable(value):
                            results[offset] = [r.target_or(r) for r in value]
                        else:
                            results[offset] = value.target_or(value)
                    return results

        def kwargs(self, pass_mode: "Recipe.PassMode") -> dict[str, Any]:
            match pass_mode:
                case Recipe.PassMode.NORMAL:
                    return self._kwargs
                case Recipe.PassMode.RESULTS:
                    results = {**self._kwargs}
                    for key in self._kwarg_keys:
                        value = results[key]
                        if is_iterable(value):
                            results[key] = [r.result() for r in value]
                        else:
                            results[key] = value.result()
                    return results
                case Recipe.PassMode.TARGETS:
                    results = {**self._kwargs}
                    for key in self._kwarg_keys:
                        value = results[key]
                        if is_iterable(value):
                            results[key] = [r.target_or(r) for r in value]
                        else:
                            results[key] = value.target_or(value)
                    return results

        def paths(self):
            return self._paths

        def component_list(self) -> ListedComponents:
            return [self._args[x] for x in self._arg_offsets]

        def component_map(self) -> MappedComponents:
            return {k: self._kwargs[k] for k in self._kwarg_keys}

        def _flat_listed_components(self) -> Generator["Recipe", None, None]:
            for c in self.component_list():
                if isinstance(c, Recipe):
                    yield c
                else:
                    yield from c

        def _flat_mapped_components(self) -> Generator["Recipe", None, None]:
            for c in self.component_map().values():
                if isinstance(c, Recipe):
                    yield c
                else:
                    yield from c

        def components(self) -> Generator["Recipe", None, None]:
            yield from self._flat_listed_components()
            yield from self._flat_mapped_components()

    @classmethod
    def scan(cls, args: Iterable[Any], kwargs: dict[str, Any]) -> Scanner:
        """
        Scan the given arguments for recipes, and return a tuple of offsets
        and keys where the recipes or recipe lists were found.

        Iterators are copied and instances of generators are expanded.

        Returns the scan results, consisting of a new args, kwargs, and lists
        of offsets and keys where recipes or recipe lists can be found.
        """

        scanner = Recipe.Scanner()

        for i, arg in enumerate(args):
            scanner.scan(arg, offset=i)

        for k, arg in kwargs.items():
            scanner.scan(arg, key=k)

        return scanner

    @classmethod
    def flat(
        cls, recipes: Iterable["Recipe"], visited: Optional[set["Recipe"]] = None
    ) -> Generator["Recipe", None, None]:
        visited = visited or set()
        for recipe in recipes:
            if recipe not in visited:
                visited.add(recipe)
                yield recipe
            yield from cls.flat(recipe.children, visited)

    def __init__(
        self,
        component_list: ComponentIterable = [],
        component_map: MappedComponents = {},
        *,
        as_user: Optional[str] = None,
        deps: Iterable["Recipe"] = [],
        docs: Optional[str] = None,
        fmt: Format = Format(),
        keep=False,
        memoize=False,
        name="(nameless)",
        parent: Optional["Recipe"] = None,
        setup: Optional["Recipe"] = None,
        sigil: Optional[FormatF] = None,
        static_files: Iterable[PathSpec] = [],
        cleanup_files: Iterable[PathSpec] = [],
        sync=False,
        target: Optional[PathSpec] = None,
    ):
        self.component_list = [*component_list]
        self.component_map = component_map
        self.id = uuid.uuid4()

        self._callsign = ""
        self._children: list["Recipe"] = []
        self._deps = [*deps]
        self._parent: Optional["Recipe"] = None
        self._target = None if target is None else Path(target)

        self.as_user = as_user
        self.docs = docs
        self.fmt = fmt
        self.keep = keep
        self.memoize = memoize
        self.name = name
        self.setup = setup
        self.static_files = [
            Path(s) for s in static_files if target is None or Path(s) != Path(target)
        ]
        self.cleanup_files = [
            Path(s) for s in cleanup_files if target is None or Path(s) != Path(target)
        ]
        self.sync = sync

        if parent:
            self.parent = parent

        if sigil:
            self.fmt = Recipe.FormatOverride(self.fmt, sigil=sigil)

        self.lock = asyncio.Lock()
        self.saved_result = None

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return f"<{self.callsign}>"

    def __iter__(self):
        yield from self.components()

    def arg(self, name: str | int):
        if isinstance(name, int):
            return self.component_list[name]
        elif isinstance(name, str):
            return self.component_map[name]
        else:
            raise ValueError(f"Invalid arg type: {type(name)}")

    @property
    def target(self) -> Path:
        assert self._target is not None, "There is no target."
        return self._target

    @target.setter
    def target(self, path: Path):
        self._target = path

    def has_target(self):
        return self._target is not None

    @property
    def parent(self) -> "Recipe":
        assert self._parent, "Recipe has no parent."
        return self._parent

    @parent.setter
    def parent(self, recipe: "Recipe"):
        self._parent = recipe
        if self not in self._parent._children:
            self._parent._children.append(self)

    @property
    def children(self) -> Iterable["Recipe"]:
        return [*self._children]

    @property
    def callsign(self) -> str:
        if not self._callsign:
            return self.sigil()
        return self._callsign

    @callsign.setter
    def callsign(self, sign: str):
        self._callsign = sign

    def has_parent(self):
        return self._parent is not None

    def sigil(self) -> str:
        return self.fmt.sigil(self)

    def log(self, event_name: str, data: Any = None):
        bus = EventBus.get()
        event = Event(event_name, self, data)
        bus.send(event)

    def error(self, msg) -> RuntimeError:
        exc = RuntimeError(msg)
        self.log(Events.ERROR, exc)
        return exc

    def has_components(self):
        try:
            next(self.components())
            return True

        except StopIteration:
            return False

    def has_dependencies(self):
        try:
            next(self.dependencies())
            return True
        except StopIteration:
            return False

    def components(self) -> Generator["Recipe", None, None]:
        yield from self.component_list
        for c in self.component_map.values():
            if isinstance(c, Recipe):
                yield c
            else:
                yield from c

    def composite_error(self, exceptions: Iterable[Exception], msg: str):
        exc = CompositeError(exceptions, msg)
        self.log(Events.ERROR, exc)
        return exc

    def age(self, ref: datetime) -> timedelta:
        if self.has_target() and self.target.exists():
            return ref - datetime.fromtimestamp(self.target.stat().st_mtime)
        if self.has_target() and self.target.is_symlink():
            return timedelta.min
        return timedelta.min

    def static_files_age(self, ref: datetime) -> timedelta:
        if not self.static_files:
            return timedelta.max
        else:
            return min(
                ref - datetime.fromtimestamp(f.stat().st_mtime)
                for f in self.static_files
            )

    def components_age(self, ref: datetime) -> timedelta:
        if not self.has_components():
            return timedelta.max
        else:
            return min(max(c.age(ref), c.inputs_age(ref)) for c in self.components())

    def add_dependency(self, dep: "Recipe"):
        assert isinstance(dep, Recipe), f"Value `{dep}` is not a recipe."
        self._deps.append(dep)
        for component in self.components():
            component.add_dependency(dep)

    def dependencies(
        self, visited_in: Optional[set["Recipe"]] = None
    ) -> Iterable["Recipe"]:
        visited = visited_in or set()

        for dep in self._deps:
            if dep not in visited:
                yield dep
                visited.add(dep)
                yield from dep.dependencies(visited)

        if self.has_parent():
            yield from self.parent.dependencies(visited)

    def dependencies_age(self, ref: datetime) -> timedelta:
        if not self.has_dependencies():
            return timedelta.max
        else:
            return min(
                max(dep.age(ref), dep.inputs_age(ref)) for dep in self.dependencies()
            )

    def inputs_age(self, ref: datetime) -> timedelta:
        return min(
            self.static_files_age(ref),
            self.components_age(ref),
            self.dependencies_age(ref),
        )

    def components_results(self) -> tuple[list[Any], dict[str, Any]]:
        results = [c.result() for c in self.component_list]
        mapped_results = {}

        for k, c in self.component_map.items():
            if isinstance(c, Recipe):
                mapped_results[k] = c.result()
            else:
                mapped_results[k] = [r.result() for r in c]

        return results, mapped_results

    def done(self) -> bool:
        if self.has_target():
            if self.target.is_symlink():
                return True
            return self.target.exists() and not self.outdated(datetime.now())
        return self.saved_result is not None

    def components_done(self) -> bool:
        return all(c.done() for c in self.components())

    def outdated(self, ref: datetime) -> bool:
        if self.has_target() and self.target.is_symlink():
            return False
        return self.age(ref) > self.inputs_age(ref)

    async def clean(self):
        remove_paths(*self.cleanup_files, as_user=self.as_user)

        if (
            not self.has_target()
            or (not self.target.exists() and not self.target.is_symlink())
            or self.keep
        ):
            return

        try:
            remove_paths(self.target, as_user=self.as_user)

        except Exception as e:
            raise self.error("Failed to clean target.") from e

        self.saved_result = None
        self.log(Events.CLEAN, self.target)

    async def clean_components(self, recursive=False):
        recipes = [*self.components()]

        if recursive:
            recipes.extend(self.dependencies())
            results = await asyncio.gather(
                *[c.clean() for c in recipes],
                *[c.clean_components(recursive=True) for c in recipes if not c.keep],
            )
        else:
            results = await asyncio.gather(
                *(c.clean() for c in recipes), return_exceptions=True
            )
        exceptions = [e for e in results if isinstance(e, Exception)]
        if exceptions:
            raise self.composite_error(
                exceptions, "Failed to clean one or more components."
            )

    async def make_dependencies(self):
        recipes = [*self.dependencies()]
        for dep in self.dependencies():
            recipes.extend(dep.dependencies())

        if self.sync:
            for c in recipes:
                try:
                    await c()
                except Exception as e:
                    raise self.error("Failed to make a dependency.") from e
        else:
            results = await asyncio.gather(
                *(c() for c in recipes), return_exceptions=True
            )
            exceptions = [e for e in results if isinstance(e, Exception)]
            if exceptions:
                raise self.composite_error(
                    exceptions, "Failed to make one or more dependencies."
                )

    async def make_components(self) -> tuple[list[Any], dict[str, Any]]:
        recipes = [*self.components()]
        if self.sync:
            results = []
            for c in recipes:
                try:
                    results.append(await c())

                except Exception as e:
                    raise self.error("Failed to make component.") from e

            return self.components_results()

        else:
            results = await asyncio.gather(
                *(c() for c in recipes), return_exceptions=True
            )
            exceptions = [e for e in results if isinstance(e, Exception)]
            if exceptions:
                raise self.composite_error(
                    exceptions, "Failed to make one or more components."
                )
            return self.components_results()

    async def make(self):
        return [r.result() for r in self.components()]

    def result(self):
        if self.done() and self.has_target():
            return self.target
        if self.saved_result is None:
            raise ValueError("Recipe result has not yet been recorded.")
        return self.saved_result

    def result_or(self, other):
        try:
            return self.result()
        except ValueError:
            return other

    def rel_target(self):
        try:
            return self.target.relative_to(Path.cwd())
        except ValueError:
            return self.target

    def rel_target_or(self, other):
        if self._target is None:
            return other
        return self.rel_target()

    def exe_target(self) -> str:
        rel_target = self.rel_target()
        if not rel_target.is_absolute() and not rel_target.parent:
            target = Path.cwd() / rel_target
            if target.is_file() and os.access(target, os.X_OK):
                return f"./{str(self.rel_target)}"
        return str(rel_target)

    def target_or(self, other):
        if self._target is None:
            return other
        return self.target

    async def _resolve(self):
        await self.make_dependencies()

        self.log(Events.START)
        await self.make_components()
        result = await self.make()

        if is_iterable(result):
            scanner = Recipe.Scanner()
            scanner.scan_params(*result)
            while scanner.has_recipes():
                await scanner.gather_all()
            result = scanner.args(pass_mode=Recipe.PassMode.RESULTS)

        elif isinstance(result, Recipe):
            result = await result()

        if self.has_target():
            assert isinstance(
                result, str | Path
            ), "Recipe declared a file target, but the result was not a string or Path object."

            result = Path(result)
            assert (
                result.resolve() == self.target.resolve()
            ), f"Recipe declared a file target, but the result path differs: {result} != {self.target}."

        self.saved_result = result

        if not self.done():
            self.saved_result = None
            raise self.error("Recipe make() didn't complete successfully.")

        return result

    async def __call__(self):
        async with self.lock:
            if self.done():
                if self.has_target():
                    return self.target
                elif self.memoize:
                    return self.saved_result

            try:
                Recipe.active.add(self)

                if self.setup is not None:
                    await self.setup()

                result = await self._resolve()
                self.log(Events.SUCCESS)
                return result

            except Exception as e:
                self.log(Events.FAIL, e)
                raise BuildError(self, str(e)) from e

            finally:
                Recipe.active.remove(self)


# --------------------------------------------------------------------
class Lambda(Recipe):
    def __init__(
        self,
        f: Callable,
        lambda_args: list[Any],
        lambda_kwargs: dict[str, Any],
        docs: Optional[str] = None,
        pass_mode=Recipe.PassMode.RESULTS,
        target_param=Recipe.DEFAULT_TARGET_PARAM,
        **kwargs,
    ):
        if "name" not in kwargs:
            kwargs = {**kwargs, "name": f.__name__}

        scanner = Recipe.scan(lambda_args, lambda_kwargs)
        sig = inspect.signature(f)
        self.bound_args = sig.bind(*lambda_args, **lambda_kwargs)
        target = self.bound_args.arguments.get(target_param, None)

        super().__init__(
            scanner.component_list(),
            scanner.component_map(),
            static_files=scanner.paths(),
            target=target,
            docs=docs or f.__doc__,
            **kwargs,
        )
        self.f = f
        self.pass_mode = pass_mode
        self.scanner = scanner

    def arg(self, name: str | int):
        if isinstance(name, str):
            return self.bound_args.arguments.get(name, None)
        else:
            return [*self.bound_args.arguments.values()][name]

    async def make(self):
        return await async_wrap(
            self.f,
            *self.scanner.args(self.pass_mode),
            **self.scanner.kwargs(self.pass_mode),
        )


# --------------------------------------------------------------------
def _inject_dependency(recipe: Recipe, dep):
    if is_iterable(dep):
        for d in dep:
            _inject_dependency(recipe, d)
    else:
        recipe.add_dependency(dep)


# --------------------------------------------------------------------
def _inject_dependencies(
    recipe: Recipe, deps: Iterable[str], bound_args: inspect.BoundArguments
):
    for dep in deps:
        dobj = bound_args.arguments.get(dep)
        _inject_dependency(recipe, dobj)


# --------------------------------------------------------------------
def recipe(
    name_or_f: Optional[str | Callable] = None,
    *,
    dep: Optional[str | Iterable[str]] = None,
    docs: Optional[str] = None,
    factory=False,
    fmt: Optional[Recipe.Format] = None,
    keep=False,
    memoize=False,
    sigil: Optional[FormatF] = None,
    sync=False,
    cleanup: Optional[str | Iterable[str]] = None,
):
    """
    Decorator for a function that defines a recipe template.

    The function is meant to be a recipe implementation method.  The
    parameters eventually passed to the method depend on whether the
    parameters are recipes or plain values.  Each recipe parameter has its
    result passed, whereas plain values are passed through unmodified.

    Can be called with no parameters.  In this mode, the name is assumed
    to be the name of the decorated function and all other parameters
    are set to their defaults.

    If `factory` is true, the function is a recipe factory that returns one
    or more recipes and the values passed to it are the recipe objects
    named.

    If `name` is provided, it is used as the name of the recipe.  Otherwise,
    the name of the recipe is inferred to be the name of the decorated
    function.

    Otherwise, the function is interpreted as a recipe implementation
    method and the values passed to it when it is eventually called
    are the result values of its dependencies.

    If `sync` is provided, the resulting recipe's dependencies are resolved
    synchronously.  Otherwise, they are resolved asynchronously using
    asyncio.gather().

    If `memoize` is provided, the recipe result is not recalculated by
    other dependencies, and the recipe implementation will only be
    evaluated once.

    If `cleanup` is provided, the referenced path(s) will additionally
    be removed when the resulting task is cleaned.
    """

    name = None if callable(name_or_f) else name_or_f

    def wrapper(f):
        @MethodAttributes.wraps(f)
        def target_wrapper(*args, **kwargs):
            truename = name or f.__name__
            scanner = Recipe.scan(args, kwargs)
            cleanup_files = [] if cleanup is None else list_or_delim(cleanup)
            cleanup_paths = [Path(s) for s in cleanup_files]

            if factory:
                if inspect.iscoroutinefunction(f):
                    raise ValueError(
                        "Recipe factories should not be coroutines.  You should define asynchronous behavior in recipes and return these from recipe factories and target definitions instead."
                    )

                result = f(
                    *scanner.args(Recipe.PassMode.NORMAL),
                    **scanner.kwargs(Recipe.PassMode.NORMAL),
                )

                if is_iterable(result):
                    result = Recipe(
                        [*result],
                        docs=docs,
                        fmt=fmt or Recipe.Format(),
                        keep=keep,
                        memoize=memoize,
                        name=truename,
                        sync=sync,
                        cleanup_files=cleanup_paths,
                    )
                elif result is None:
                    result = Recipe([], docs=docs, name=truename)
                else:
                    result = cast(Recipe, result)

                    result.docs = docs
                    result.fmt = fmt or result.fmt
                    result.keep = keep
                    result.memoize = memoize
                    result.name = truename
                    result.sync = sync
                    result.cleanup_files = cleanup_paths

            else:
                result = Lambda(
                    f,
                    [*args],
                    {**kwargs},
                    docs=docs,
                    fmt=fmt or Recipe.Format(),
                    keep=keep,
                    memoize=memoize,
                    name=truename,
                    sync=sync,
                    cleanup_files=cleanup_paths,
                )

            if sigil:
                result.fmt = Recipe.FormatOverride(result.fmt, sigil=sigil)

            _inject_dependencies(
                result,
                [] if dep is None else list_or_delim(dep),
                scanner.bind(f, Recipe.PassMode.NORMAL),
            )

            return result

        return target_wrapper

    if callable(name_or_f):
        return no_type_check(wrapper(name_or_f))

    return no_type_check(cast(Callable[[Callable], Callable], wrapper))



# --------------------------------------------------------------------
# abstract_injector.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------


from abc import ABCMeta, abstractmethod
from typing import Any, List, Optional, Set

from .attributes import NOTHING, ClassAttributes, get_providers
from .decorators import Tags
from .errors import (
    CircularDependencyError,
    InjectionError,
    InvalidResourceError,
    MissingDependencyError,
    MissingResourceError,
)
from .namespaces import Namespace


# --------------------------------------------------------------------
class AbstractInjector(metaclass=ABCMeta):
    """
    This is the abstract base class for injectors.

    An injector is responsible for collecting resources from modules and
    injecting them into newly created instances and/or providing them
    when explicitly required.

    An Injector is instantianted with any number of modules, which
    provide resources via methods annotated with @provide.  These
    methods are allowed to depend on other resources, potentially
    from myriad other modules, in order to generate their
    dependencies.

    Evaulation and injection of dependencies is lazy, meaning that
    the dependency graph is not evaluated until a resource is
    requested.
    """

    def __init__(self, *modules):
        """
        Create an Injector object.

        *modules: A list of modules to include in the injector.
                  More modules can be added later by calling
                  Injector.add_module().
        """
        self.resources: dict[str, Any] = {"injector": lambda: self}
        self.singletons = {}
        self.dep_graph = {}
        self.injection_interceptors = []
        self.ns_index = Namespace.root()
        self.resource_attrs = {}

        for module in modules:
            self.add_module(module, skip_cycle_check=True)
        self.check_for_cycles()

    @abstractmethod
    def create(self, class_):
        """
        Create an instance of the specified class.  The class' constructor
        must follow the rules for @inject methods, such that all of its
        parameters refer to injectable resources or are optional.

        If the object needs to be constructed with objects not from the
        Injector, do not use this method.  Instead, instantiate the object
        with these parameters in the constructor, then mark one or more
        methods with @inject and pass the instance to Injector.inject().
        """
        raise NotImplementedError()

    @abstractmethod
    def inject(self, obj, aliases={}, namespace=""):
        """
        Inject a method or object instance with resources from this Injector.

        obj: A method or object instance.  If this is a method, all named
             parameters are injected from the Injector.  If this is an
             instance, its methods are scanned for injection points and these
             methods are all invoked with resources from the Injector.
        aliases: An optional map from dependency alias to real dependency name.
        """
        raise NotImplementedError()

    @abstractmethod
    def require(self, name, method=None):
        """
        Require a named resource from this Injector.  If it can't be provided,
        an InjectionError is raised.

        This method should only be called from outside of the asyncio
        event loop.

        Optionally, a method can be specified.  This indicates the method that
        requires the resource for injection, and causes this method to throw
        MethodInjectionError instead of InjectionError if the resource was
        not provided.
        """
        raise NotImplementedError()

    @abstractmethod
    def provide(
        self, name_or_method, value=NOTHING, is_singleton=False, namespace=None
    ):
        raise NotImplementedError()

    @abstractmethod
    def _bind_resource(self, bound_method, module_aliases={}, namespace=None):
        raise NotImplementedError()

    def get_namespace(self, name=None) -> Optional[Namespace]:
        if name is None or name == Namespace.SEP:
            return self.ns_index
        return self.ns_index.get_namespace(name)

    def add_module(self, module, skip_cycle_check=False):
        """
        Add a module to the injector.  The module is scanned for @provider
        annotated methods, and these methods are added as resources to
        the injector.
        """
        module_attrs = ClassAttributes.for_class(module.__class__)
        assert module_attrs is not None
        namespace = module_attrs.get(Tags.NAMESPACE, None)
        using_namespaces = []
        if namespace is not None:
            using_namespaces.append(namespace)
            self.ns_index.add_namespace(namespace)
        module_aliases = self._get_aliases(module_attrs, using_namespaces)
        for name, value in module_attrs.get(Tags.CONST_MAP, {}).items():
            self.provide(name, value, is_singleton=True, namespace=namespace)
        for _, provider in get_providers(module):
            self._bind_resource(provider, module_aliases, namespace)

        if not skip_cycle_check:
            self.check_for_cycles()

    def add_injection_interceptor(self, interceptor):
        """
        Specifies a function to be called before resources are injected into
        a provider, constructor, or injection point.  Resources are provided
        to the function as a key/value map.

        The injection interceptor is expected to return a key/value map
        containing all of the resources provided to it, either modified
        or in their original form.  Failure to provide all required
        resources will lead to an InjectionError after the interceptors
        are invoked.
        """
        self.injection_interceptors.append(interceptor)

    def has(self, name):
        """
        Determine if this Injector has been provided with the named resource.
        """
        return name in self.resources

    def get_dependency_tree(self, resource_name):
        if resource_name not in self.resources:
            raise MissingResourceError(resource_name)
        try:
            return {
                dep: self.get_dependency_tree(dep)
                for dep in self.get_dependencies(resource_name)
            }
        except MissingResourceError as e:
            raise MissingDependencyError(resource_name, e.name)

    def get_dependency_graph(self, resource_name, *other_resource_names):
        resource_names = [resource_name, *other_resource_names]
        for name in resource_names:
            if name not in self.resources:
                raise MissingResourceError(name)
        try:
            dep_graph = {name: self.get_dependencies(name) for name in resource_names}
            for dep in dep_graph[resource_name]:
                dep_graph.update(self.get_dependency_graph(dep))
            return dep_graph
        except MissingResourceError as e:
            raise MissingDependencyError(resource_name, e.name)

    def get_ordered_dependencies(self, resource_name):
        deps: List[str] = []
        visited: Set[str] = set()
        stack: List[str] = [resource_name]

        while stack:
            name = stack.pop()
            if name not in visited:
                visited.add(name)
                if not name == resource_name:
                    deps.append(name)
                stack.extend(self.get_dependencies(name))

        return deps

    def get_dependencies(self, resource_name):
        return self.dep_graph.get(resource_name, lambda: ())()

    def get_resource_attributes(self, resource_name):
        if resource_name not in self.resource_attrs:
            raise MissingResourceError(resource_name)
        return self.resource_attrs[resource_name]

    def scan_resources(self, filter_f):
        for key, value in self.resource_attrs.items():
            if filter_f(key, value):
                yield key, value

    def unbind_singleton(self, resource_name=None, unbind_all=False):
        if unbind_all:
            self.singletons.clear()
        else:
            if resource_name not in self.resources:
                raise MissingResourceError(resource_name)
            if not self.get_resource_attributes(resource_name).check(Tags.SINGLETON):
                raise InvalidResourceError(
                    'Resource "%s" is not a singleton.' % resource_name
                )
            if resource_name in self.singletons:
                del self.singletons[resource_name]

    def check_for_cycles(self):
        def visit(resource, visited=None):
            if visited is None:
                visited = set()
            visited.add(resource)
            for dep in self.dep_graph.get(resource, lambda: ())():
                if dep in visited:
                    raise CircularDependencyError(resource, dep)
                visit(dep, set(visited))

        for resource in self.dep_graph:
            visit(resource)

    def _get_aliases(self, attrs, namespaces=[]):
        aliases = attrs.get(Tags.ALIASES, {})
        for alias in aliases.keys():
            if Namespace.SEP in alias:
                raise InjectionError(
                    "Alias name may not contain the namespace "
                    'separator: "%s"' % alias
                )
        using_namespaces = namespaces + attrs.get(Tags.USING_NAMESPACES, [])
        for ns_name in using_namespaces:
            namespace = self.get_namespace(ns_name)
            assert namespace is not None
            aliases = {
                **aliases,
                **{
                    Namespace.leaf_name(name): Namespace.join(ns_name, name)
                    for name in namespace.get_leaves()
                },
            }
        return aliases

    def _invoke_injection_interceptors(self, attrs, param_map, alias_map):
        for interceptor in self.injection_interceptors:
            param_map = interceptor(attrs, param_map, alias_map)
        return param_map



# --------------------------------------------------------------------
# cookbook.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Tuesday April 4, 2023
#
# Distributed under terms of the MIT license.
#
# NOTE: This module is deprecated.  The functions and classes that used
# to be in this module have been moved to other modules.  Please do not
# add further functions or classes to this module.  The ones imported
# below are included for backwards compatibility with older builds.
#
# This file will be removed in the next major version (v7.x).
# --------------------------------------------------------------------

from xeno.recipe import recipe
from xeno.recipes.shell import ShellRecipe, sh



# --------------------------------------------------------------------
# color.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Wednesday October 28, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------
import io
import os
import sys
from functools import partial
from typing import List, Optional

# --------------------------------------------------------------------
_ansi_enabled = (
    "NO_COLOR" not in os.environ and sys.stdout.isatty()
) or "FORCE_COLOR" in os.environ


# --------------------------------------------------------------------
def disable():
    global _ansi_enabled
    _ansi_enabled = False


# --------------------------------------------------------------------
def enable():
    global _ansi_enabled
    _ansi_enabled = True


# --------------------------------------------------------------------
def is_enabled():
    global _ansi_enabled
    return _ansi_enabled


# --------------------------------------------------------------------
COLORS = ("black", "red", "green", "yellow", "blue", "magenta", "cyan", "white")
RENDER_MODES = (
    "none",
    "bold",
    "dim",
    "italic",
    "underline",
    "blink",
    "blink2",
    "negative",
    "concealed",
    "crossed",
)

# --------------------------------------------------------------------
ANSI_ESC = "\033["


# --------------------------------------------------------------------
def seq(code) -> str:
    return ANSI_ESC + str(code)


# --------------------------------------------------------------------
def attr(*parts) -> str:
    return seq(";".join(str(p) for p in parts)) + "m"


# --------------------------------------------------------------------
def clreol():
    if sys.stdout.isatty():
        sys.stdout.write(seq("K"))
        sys.stdout.flush()


# --------------------------------------------------------------------
def show_cursor():
    if sys.stdout.isatty():
        sys.stdout.write(seq("?25h"))
        sys.stdout.flush()


# --------------------------------------------------------------------
def hide_cursor():
    if sys.stdout.isatty():
        sys.stdout.write(seq("?25l"))


# --------------------------------------------------------------------
RESET = attr(0)


# --------------------------------------------------------------------
def style(
    fg: Optional[str] = None, bg: Optional[str] = None, render: Optional[str] = None
):
    attr_codes: List[int] = []

    if render is not None:
        attr_codes.append(RENDER_MODES.index(render))
    if fg is not None:
        attr_codes.append(30 + COLORS.index(fg))
    if bg is not None:
        attr_codes.append(40 + COLORS.index(bg))

    return attr(*attr_codes)


# --------------------------------------------------------------------
def color(
    *content,
    fg: Optional[str] = None,
    bg: Optional[str] = None,
    render: Optional[str] = None,
    after: str = ""
) -> str:
    if _ansi_enabled:
        before = style(fg, bg, render)
        return before + "".join(str(obj) for obj in content) + RESET + after
    return "".join(str(obj) for obj in content)


# --------------------------------------------------------------------
class TextDecorator:
    def __init__(
        self,
        *,
        fg: Optional[str] = None,
        bg: Optional[str] = None,
        render: Optional[str] = None,
        outfile=sys.stdout
    ):
        self.fg = fg
        self.bg = bg
        self.render = render
        self.outfile = outfile
        self.wipe = 0

    def embrace(
        self,
        text,
        *,
        begin="[",
        end="]",
        brace_fg="white",
        brace_bg=None,
        brace_render="bold",
        **kwargs
    ):
        self._autowipe()
        brace_color = partial(color, fg=brace_fg, bg=brace_bg, render=brace_render)
        sb = io.StringIO()
        sb.write(brace_color(begin))
        sb.write(color(text, **kwargs))
        sb.write(brace_color(end))
        return self.outfile.write(sb.getvalue() + " ")

    def write(self, text, **kwargs):
        return self.outfile.write(color(text, **self._inject_kwargs(kwargs)))

    def print(self, text, **kwargs):
        self._autowipe()
        n = self.write(text, **kwargs)
        n += self.outfile.write("\n")
        self.flush()
        return n

    def flush(self):
        return self.outfile.flush()

    def wipeline(self, n: int):
        if n > 0:
            self.outfile.write("\r")
            self.outfile.write(" " * n)
            self.outfile.write("\r")
            self.flush()

    def _inject_kwargs(self, kwargs):
        return {
            **kwargs,
            "fg": self.fg or kwargs.get("fg", None),
            "bg": self.bg or kwargs.get("bg", None),
            "render": self.render or kwargs.get("render", None),
        }

    def _autowipe(self):
        self.wipeline(self.wipe)
        self.wipe = 0

    def __call__(self, text, **kwargs):
        return self.print(text, **kwargs)



# --------------------------------------------------------------------
# decorators.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import inspect

from .attributes import ClassAttributes, MethodAttributes, Tags

# --------------------------------------------------------------------
def singleton(f):
    """
    Method annotation indicating a named singleton resource.

    The function will only ever be invoked on an instance of the
    module once, and the return value will be provided to all
    injected objects that require it.
    """

    attrs = MethodAttributes.for_method(f, write=True)
    assert attrs is not None
    attrs.put(Tags.SINGLETON)
    attrs.put(Tags.PROVIDER)
    return f


# --------------------------------------------------------------------
def provide(f):
    """
    Method annotation indicating a named resource.

    The function will be added to the Injector's resource map and
    called each time an injected instance is created that requires
    the resource.
    """

    attrs = MethodAttributes.for_method(f, write=True)
    assert attrs is not None
    attrs.put(Tags.PROVIDER)
    return f


# --------------------------------------------------------------------
def inject(f):
    """
    Method annotation indicating an injection point in an object.

    Instance methods marked with @inject are called after an object
    is created via Injector.create() or if an instance is passed
    to Injector.inject().

    All of the parameters of a method marked with @inject must refer
    to named resources in the Injector, or must provide default values
    which can be overridden by resources in the injector.
    """
    attrs = MethodAttributes.for_method(f, write=True)
    assert attrs is not None
    attrs.put(Tags.INJECTION_POINT)
    return f


# --------------------------------------------------------------------
def named(name):
    """
    Method annotation indicating a name for the given resource other
    than the name of the method itself.
    """

    def impl(f):
        attrs = MethodAttributes.for_method(f, write=True)
        assert attrs is not None
        attrs.put(Tags.NAME, name)
        return f

    return impl


# --------------------------------------------------------------------
def alias(alias, name):
    """
    Aliases a single resource to a different name in the given
    module or resource context.
    """

    def impl(obj):
        attrs = None
        if inspect.isclass(obj):
            attrs = ClassAttributes.for_class(obj, write=True)
        else:
            attrs = MethodAttributes.for_method(obj, write=True)
        assert attrs is not None
        aliases = attrs.get(Tags.ALIASES, {})
        aliases[alias] = name
        attrs.put(Tags.ALIASES, aliases)
        return obj

    return impl


# --------------------------------------------------------------------
def namespace(name):
    """
    Module annotation indicating that the resources defined inside
    should be scoped into the given namespace, that is the given
    string is appended to all resource names followed by '/'.
    """

    def impl(class_):
        attrs = ClassAttributes.for_class(class_, write=True)
        assert attrs is not None
        attrs.put(Tags.NAMESPACE, name)
        return class_

    return impl


# --------------------------------------------------------------------
def const(name, value):
    """
    Module annotation defining a constant resource scoped into the
    module's namespace.
    """

    def impl(class_):
        attrs = ClassAttributes.for_class(class_, write=True)
        assert attrs is not None
        const_map = attrs.get(Tags.CONST_MAP, {})
        const_map[name] = value
        attrs.put(Tags.CONST_MAP, const_map)
        return class_

    return impl


# --------------------------------------------------------------------
def using(name):
    def impl(obj):
        attrs = None
        if inspect.isclass(obj):
            attrs = ClassAttributes.for_class(obj, write=True)
        else:
            attrs = MethodAttributes.for_method(obj, write=True)
        assert attrs is not None
        namespaces = attrs.get(Tags.USING_NAMESPACES, [])
        namespaces.append(name)
        attrs.put(Tags.USING_NAMESPACES, namespaces)
        return obj

    return impl



#!/usr/bin/env python
# --------------------------------------------------------------------
# recipes/__init__.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Monday August 28, 2023
#
# Distributed under terms of the MIT license.
# -------------------------------------------------------------------

import ast
from pathlib import Path
from typing import Collection, Optional

from xeno.recipe import Recipe, recipe
from xeno.recipes.shell import sh
from xeno.shell import select_env
from xeno.typedefs import PathSpec

# -------------------------------------------------------------------
INSTALL_ENV = select_env(append="PREFIX,DESTDIR", PREFIX="usr/local", DESTDIR="/")
TEST_ENV = select_env()


# -------------------------------------------------------------------
@recipe(factory=True, sigil=lambda r: f"{r.name}:{r.target.name}")
def install(program: Recipe, env=INSTALL_ENV):
    root = env.get("DESTDIR", "/")
    prefix = env.get("PREFIX", "/usr/local")

    path = Path(root) / prefix / "bin" / program.target.name
    return sh(
        "mkdir -p {DESTDIR}{PREFIX}/bin; "
        "cp -f {program} {target}; "
        "chmod 775 {target}",
        env=env,
        program=program,
        target=path,
        as_user="root",
    )


# -------------------------------------------------------------------
@recipe(factory=True, sigil=lambda r: f"{r.name}:{r.arg('t').target.name}")
def test(t, env=TEST_ENV, interactive: Collection[str] = set()):
    result = sh(
        t.target,
        t=t,
        env=env,
        interactive=t.name in interactive,
    )
    return result


# -------------------------------------------------------------------
@recipe
async def repo_deps(repo):
    build_script = Path(repo) / "build.py"
    if not build_script.exists():
        return 0

    with open(build_script, "r") as infile:
        tree = ast.parse(infile.read())

    def has_xeno_imports():
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    if alias.name.startswith("xeno"):
                        return True
            elif isinstance(node, ast.ImportFrom):
                if node.module and node.module.startswith("xeno"):
                    return True
        return False

    def has_deps_target():
        for node in ast.walk(tree):
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                if node.name == "deps":
                    return True
        return False

    if has_xeno_imports() and has_deps_target():
        return await sh("./build.py deps", repo=repo, cwd=repo, check=True).make()

    return 0


# -------------------------------------------------------------------
@recipe(
    factory=True, sigil=lambda r: f"{r.name}:{r.arg('repo').target.name.split('/')[-1]}"
)
def checkout(repo, target: Optional[PathSpec] = None):
    name = repo.split("/")[-1]
    target = target or Path("deps") / name
    return repo_deps(
        sh(["git", "clone", "{repo}", "{target}"], repo=repo, target=target)
    )



# --------------------------------------------------------------------
# util.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import inspect
from collections.abc import Sequence
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Generator, Iterable, Union, TypeVar

from xeno.errors import InjectionError
from xeno.typedefs import NestedIterable


# --------------------------------------------------------------------
T = TypeVar("T")


# --------------------------------------------------------------------
async def async_map(key, coro):
    """
    Wraps a coroutine so that when executed, the coroutine result
    and the mapped value are provided.  Useful for gathering results
    from a map of coroutines.
    """
    return key, await coro


# --------------------------------------------------------------------
async def async_wrap(f, *args, **kwargs):
    """
    Wraps a normal function in a coroutine.  If the given function
    is already a coroutine function, we simply yield from it.
    """
    if not asyncio.iscoroutinefunction(f):
        return f(*args, **kwargs)
    return await f(*args, **kwargs)


# --------------------------------------------------------------------
async def async_vwrap(v):
    """
    Wraps a value in a coroutine for gathering.
    """
    return v


# --------------------------------------------------------------------
def bind_unbound_method(obj, method):
    return method.__get__(obj, obj.__class__)


# --------------------------------------------------------------------
def decode(b: bytes) -> str:
    try:
        return b.decode("utf-8")
    except UnicodeDecodeError:
        return b.decode("ISO-8859-1")


# --------------------------------------------------------------------
def get_params_from_signature(f):
    """
    Fetches the params tuple list from the given function's signature.
    """
    sig = inspect.signature(f)
    return list(sig.parameters.values())


# --------------------------------------------------------------------
def is_iterable(obj: Any) -> bool:
    """Determine if the given object is an iterable sequence other than a string or byte array."""
    return (
        isinstance(obj, Sequence)
        and not isinstance(obj, (str, bytes, bytearray))
        or inspect.isgenerator(obj)
    )


# --------------------------------------------------------------------
def file_age(file: Path) -> timedelta:
    return datetime.now() - datetime.fromtimestamp(file.stat().st_mtime)


# --------------------------------------------------------------------
def resolve_alias(name, aliases, visited=None):
    if visited is None:
        visited = set()

    if name in aliases:
        if name in visited:
            raise InjectionError(
                "Alias loop detected: %s -> %s" % (name, ",".join(visited))
            )
        visited.add(name)
        name = resolve_alias(aliases[name], aliases, set(visited))
    return name


# --------------------------------------------------------------------
def list_or_delim(obj: str | Iterable[str], delim=",") -> Generator[str, None, None]:
    if isinstance(obj, str):
        yield from list_or_delim(obj.split(delim))
    else:
        yield from obj


# --------------------------------------------------------------------
def expand(val: T | NestedIterable[T]) -> Iterable[T]:
    if is_iterable(val):
        for v in val:
            yield from expand(v)
    else:
        yield val



# --------------------------------------------------------------------
# events.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday March 9, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

"""
Asynchronous event dispatch and broadcast.
"""

import asyncio
import inspect
import uuid
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Callable, Generator, Optional

from xeno.utils import async_wrap


# --------------------------------------------------------------------
@dataclass
class Event:
    name: str
    context: Any = None
    data: Any = None
    when: datetime = field(default_factory=datetime.now)

    def age(self) -> timedelta:
        return datetime.now() - self.when


# --------------------------------------------------------------------
EventListener = Callable[[Event], Any]


# --------------------------------------------------------------------
@dataclass
class Timer:
    duration: timedelta
    next: datetime
    listener: EventListener
    interval: bool = False
    id: uuid.UUID = field(default_factory=uuid.uuid4)

    def __hash__(self):
        return hash(self.id)


# --------------------------------------------------------------------
class EventBus:
    FRAME = "EventBus.FRAME"
    TIMER = "EventBus.TIMER"

    _current_bus: Optional["EventBus"] = None

    class _Session:
        def __init__(self):
            self.bus = EventBus()

        def __enter__(self):
            EventBus._current_bus = self.bus

        def __exit__(self, *_):
            EventBus._current_bus = None
            self.bus.shutdown()

    @staticmethod
    def session() -> "EventBus._Session":
        return EventBus._Session()

    @staticmethod
    def get():
        if EventBus._current_bus is None:
            raise ValueError("There is no current event bus session.")
        return EventBus._current_bus

    def __init__(self):
        self.queue: asyncio.Queue[Event] = asyncio.Queue()
        self.shutdown_flag = asyncio.Event()
        self.async_listeners: set[EventListener] = set()
        self.sync_listeners: set[EventListener] = set()
        self.async_subs: defaultdict[str, set[EventListener]] = defaultdict(set)
        self.sync_subs: defaultdict[str, set[EventListener]] = defaultdict(set)
        self.timers: set[Timer] = set()

    def send(self, event: Event):
        if self._has_async_listeners_for_event(event):
            self.queue.put_nowait(event)
        self._dispatch_sync(event)

    def shutdown(self):
        self.shutdown_flag.set()
        self.sync_listeners.clear()
        self.sync_subs.clear()
        self.async_listeners.clear()
        self.async_subs.clear()
        while not self.queue.empty():
            self.queue.get_nowait()

    def set_timer(self, duration: int | timedelta, listener: EventListener) -> Timer:
        now = datetime.now()
        if isinstance(duration, int):
            duration = timedelta(seconds=duration)
        timer = Timer(duration, now + duration, listener)
        self.timers.add(timer)
        return timer

    def set_interval(self, duration: int | timedelta, listener: EventListener) -> Timer:
        timer = self.set_timer(duration, listener)
        timer.interval = True
        return timer

    def remove_timer(self, timer: Timer):
        if timer in self.timers:
            self.timers.remove(timer)

    def listen(self, listener: EventListener):
        if inspect.iscoroutinefunction(listener):
            self.async_listeners.add(listener)
        else:
            self.sync_listeners.add(listener)

    def unlisten(self, listener: EventListener):
        try:
            self.sync_listeners.remove(listener)
        except KeyError:
            pass

        try:
            self.async_listeners.remove(listener)
        except KeyError:
            pass

    def subscribe(self, event: str, listener: EventListener):
        if inspect.iscoroutinefunction(listener):
            self.async_subs[event].add(listener)
        else:
            self.sync_subs[event].add(listener)

    def unsubscribe(self, event: str, listener: EventListener):
        subs = self.sync_subs[event]
        try:
            subs.remove(listener)
        except KeyError:
            pass
        if not subs:
            del self.sync_subs[event]

        subs = self.async_subs[event]
        try:
            subs.remove(listener)
        except KeyError:
            pass
        if not subs:
            del self.async_subs[event]

    async def run(self):
        while not self.shutdown_flag.is_set():
            try:
                event = self.queue.get_nowait()
                await self._dispatch_async(event)

            except asyncio.QueueEmpty:
                pass

            await self._dispatch_timers()
            await self._dispatch_frame_event()
            await asyncio.sleep(0)

    def _has_async_listeners_for_event(self, event: Event) -> bool:
        try:
            next(self._async_listeners_for_event(event))
            return True
        except StopIteration:
            return False

    def _async_listeners_for_event(
        self, event: Event
    ) -> Generator[EventListener, None, None]:
        yield from self.async_listeners
        if event.name in self.async_subs:
            yield from self.async_subs[event.name]

    def _sync_listeners_for_event(
        self, event: Event
    ) -> Generator[EventListener, None, None]:
        yield from self.sync_listeners
        if event.name in self.sync_subs:
            yield from self.sync_subs[event.name]

    async def _dispatch_timers(self):
        elapsed_timers = []
        now = datetime.now()

        for timer in self.timers:
            if now >= timer.next:
                elapsed_timers.append(timer)

        sync_callbacks = []
        async_callbacks = []

        for timer in elapsed_timers:
            if timer.interval:
                timer.next = now + timer.duration
            else:
                self.timers.remove(timer)

            if inspect.iscoroutinefunction(timer.listener):
                async_callbacks.append((timer.listener, Event(EventBus.TIMER, timer)))
            else:
                sync_callbacks.append((timer.listener, Event(EventBus.TIMER, timer)))

        await asyncio.gather(*[
            callback(evt) for callback, evt in async_callbacks
        ])

        for callback, evt in sync_callbacks:
            callback(evt)

    async def _dispatch_frame_event(self):
        event = Event(EventBus.FRAME, self)
        await self._dispatch_async(event)
        self._dispatch_sync(event)

    def _dispatch_sync(self, event: Event):
        for listener in self._sync_listeners_for_event(event):
            listener(event)

    async def _dispatch_async(self, event: Event):
        await asyncio.gather(
            *(
                async_wrap(listener, event)
                for listener in self._async_listeners_for_event(event)
            )
        )


# --------------------------------------------------------------------
event_session = EventBus.session



# --------------------------------------------------------------------
# typedefs.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Monday August 28, 2023
# --------------------------------------------------------------------

import asyncio
from pathlib import Path
from typing import Any, Callable, Dict, Tuple, Union, TypeVar, Iterable


# --------------------------------------------------------------------
T = TypeVar("T")

# --------------------------------------------------------------------
EnvDict = Dict[str, str] | Dict[str, Any]
InputSource = Callable[[], str]
LineSink = Callable[[str, asyncio.StreamWriter], None]
OutputTaskData = Tuple[asyncio.StreamReader, LineSink]
PathSpec = Union[str | Path]
NestedIterable = Iterable[T | "NestedIterable[T]"]



# --------------------------------------------------------------------
# build.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Friday March 17, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import fnmatch
import multiprocessing
import sys
import traceback
from argparse import ArgumentParser, HelpFormatter
from collections import defaultdict
from datetime import datetime, timedelta
from typing import Any, Callable, Iterable, Optional, cast

from xeno.async_injector import AsyncInjector
from xeno.attributes import MethodAttributes
from xeno.color import TextDecorator
from xeno.color import disable as disable_color
from xeno.color import enable as enable_color
from xeno.color import is_enabled as is_color_enabled
from xeno.cookbook import recipe as base_recipe
from xeno.decorators import named
from xeno.events import Event, EventBus
from xeno.recipe import BuildError, Events, Recipe
from xeno.shell import Environment, Shell
from xeno.spinner import Spinner

# --------------------------------------------------------------------
EngineHook = Callable[["Config", "Engine", EventBus], None]


# --------------------------------------------------------------------
class Config:
    class Mode:
        BUILD = "build"
        CLEAN = "clean"
        HELP = "help"
        LIST = "list"
        LIST_ALL = "list_all"
        REBUILD = "rebuild"
        TREE = "tree"
        QUERY = "query"

    class CleanupMode:
        NONE = "none"
        SHALLOW = "shallow"
        RECURSIVE = "recursive"

    class ColorOptions:
        YES = "yes"
        NO = "no"
        AUTO = "auto"

    class SortingHelpFormatter(HelpFormatter):
        def add_arguments(self, actions):
            actions = sorted(actions, key=lambda a: a.option_strings)
            super().add_arguments(actions)

    def __init__(self):
        self.cleanup_mode = self.CleanupMode.NONE
        self.color = "auto"
        self.debug = False
        self.jobs = multiprocessing.cpu_count()
        self.mode = self.Mode.BUILD
        self.quiet = False
        self.targets: list[str] = []
        self.tasks: list[str] = []
        self.to_stdout = False
        self.verbose = 0
        self.query: Optional[str] = None

    def _argparser(self):
        parser = ArgumentParser(
            add_help=False,
            formatter_class=Config.SortingHelpFormatter,
        )
        parser.add_argument(
            "targets",
            nargs="*",
            help="Top-level task targets or addressable recipes to build.",
        )
        parser.add_argument(
            "--help",
            "-h",
            dest="mode",
            action="store_const",
            const=self.Mode.HELP,
            help="Show this help text.",
        )
        parser.add_argument(
            "--clean",
            "-c",
            dest="cleanup_mode",
            action="store_const",
            const=self.CleanupMode.RECURSIVE,
            help="Clean the specified or default targets and all their components.",
        )
        parser.add_argument(
            "--cut",
            "-x",
            dest="cleanup_mode",
            action="store_const",
            const=self.CleanupMode.SHALLOW,
            help="Clean the specified or default targets only.",
        )
        parser.add_argument(
            "--rebuild",
            "-R",
            dest="mode",
            action="store_const",
            const=self.Mode.REBUILD,
            help="Clean and then rebuild the specified or default targets and their components.",
        )
        parser.add_argument(
            "--verbose",
            "-v",
            action="count",
            help="Print additional diagnostic or help info.",
        )
        parser.add_argument(
            "--quiet",
            "-q",
            action="store_true",
            help="Only print info/warning/error event contents to stderr, nothing else.",
        )
        parser.add_argument(
            "--query",
            "-Q",
            help="Return 0 if the given target or comma separated list of targets exist, 1 otherwise, do nothing else.",
        )
        parser.add_argument(
            "--to-stdout",
            "-s",
            action="store_true",
            help="Print info event contents to stdout, all else to stderr.",
        )
        parser.add_argument(
            "--list",
            "-l",
            dest="mode",
            action="store_const",
            const=self.Mode.LIST,
            help="List all top-level tasks targets.",
        )
        parser.add_argument(
            "--list-all",
            "-L",
            dest="mode",
            action="store_const",
            const=self.Mode.LIST_ALL,
            help="List all top-level task targets and their addressable components.",
        )
        parser.add_argument(
            "--tree",
            "-T",
            dest="mode",
            action="store_const",
            const=self.Mode.TREE,
            help="List all tasks and their components in a tree.",
        )
        parser.add_argument(
            "--debug",
            "-D",
            action="store_true",
            help="Enable debug and extra diagnostic info.",
        )
        parser.add_argument(
            "--color",
            choices=["yes", "no", "auto"],
            default="auto",
            help="Choose when to enable colorized output.",
        )
        parser.add_argument(
            "--jobs",
            "-j",
            dest="jobs",
            type=int,
            help="Number of simultaneous shells, defaults to number of CPUs.",
        )
        parser.set_defaults(mode=self.Mode.BUILD)
        return parser

    def parse_args(self, *args):
        self._argparser().parse_args(args, namespace=self)
        if self.cleanup_mode != Config.CleanupMode.NONE:
            self.mode = Config.Mode.CLEAN
        if self.mode == Config.Mode.REBUILD:
            self.cleanup_mode = Config.CleanupMode.RECURSIVE
        if self.query is not None:
            self.mode = Config.Mode.QUERY
        return self


# --------------------------------------------------------------------
class Engine:
    TREE_SUB_ARROW = "⮡"

    class Attributes:
        TARGET = "xeno.build.task"
        DEFAULT = "xeno.build.default"

    def __init__(self, name="Build Script"):
        self.name = name
        self.bus_hooks: list[EngineHook] = list()
        self.env = Environment.context()
        self.injector = AsyncInjector()
        self.scan = Recipe.Scanner()
        self.txt = TextDecorator()

        self._build_mode_methods = {
            Config.Mode.BUILD: self._build_mode_build,
            Config.Mode.CLEAN: self._build_mode_clean,
            Config.Mode.HELP: self._build_mode_help,
            Config.Mode.LIST: self._build_mode_list,
            Config.Mode.QUERY: self._build_mode_query,
            Config.Mode.LIST_ALL: self._build_mode_list_all,
            Config.Mode.REBUILD: self._build_mode_rebuild,
            Config.Mode.TREE: self._build_mode_tree,
        }

    def add_hook(self, hook: EngineHook):
        self.bus_hooks.append(hook)

    async def tasks(self, *, parent: Optional[Recipe] = None) -> list[Recipe]:
        tasks = []
        if parent is None:
            root_names = [
                k
                for k, _ in self.injector.scan_resources(
                    lambda _, v: v.check(self.Attributes.TARGET)
                )
            ]

            for name in root_names:
                recipe = await self.injector.require_async(name)
                assert isinstance(
                    recipe, Recipe
                ), f"Task `{name}` did not yield a recipe."

                recipe.callsign = name
                tasks.append(recipe)
                await self.tasks(parent=recipe)

        else:
            for recipe in parent.components():
                recipe.parent = parent
                tasks.append(recipe)

        return tasks

    async def addressable_task_map(self) -> dict[str, Recipe]:
        all_tasks = [*Recipe.flat(await self.tasks())]
        callsign_counts: dict[str, int] = defaultdict(lambda: 0)
        for task in all_tasks:
            callsign_counts[task.callsign] += 1
        duplicate_callsigns = [k for k, v in callsign_counts.items() if v > 1]
        task_map = {t.callsign: t for t in all_tasks}
        for callsign in duplicate_callsigns:
            del task_map[callsign]
        return task_map

    def default_task(self) -> Optional[str]:
        results = [
            k
            for k, _ in self.injector.scan_resources(
                lambda _, v: v.check(self.Attributes.DEFAULT)
            )
        ]
        assert len(results) <= 1, "More than one default task specified."
        return results[0] if results else None

    def provide(self, *args, **kwargs):
        self.injector.provide(*args, **{**kwargs, "is_singleton": True})

    def recipe(self, *args, **kwargs):
        return base_recipe(*args, **kwargs)

    def factory(self, *args, **kwargs):
        return base_recipe(*args, **kwargs, factory=True)

    def task(
        self,
        name_or_f: Optional[str | Callable] = None,
        *,
        dep: Optional[str | Iterable[str]] = None,
        default=False,
        factory=True,
        fmt: Optional[Recipe.Format] = None,
        keep=False,
        memoize=True,
        sync=False,
        cleanup: Optional[str | Iterable[str]] = None,
    ):
        """
        Decorator for defining a task recipe for a build.

        Can be called with no parameters.  In this mode, the name is assumed
        to be the name of the decorated function and all other parameters
        are set to their defaults.

        If `default` is True, the task will be the default task
        when no task is specified at build time.  This method will
        throw ValueError if another task has already been specified
        as the default task.

        See `xeno.cookbook.recipe` for info about the other params, note that
        `factory` and `memoize` params to `xeno.cookbook.recipe()` are always
        `True` here.
        """

        name = None if callable(name_or_f) else name_or_f

        def wrapper(f):
            target_wrapper = cast(
                Recipe,
                base_recipe(
                    name,
                    dep=dep,
                    docs=f.__doc__,
                    factory=factory,
                    keep=keep,
                    sync=sync,
                    memoize=memoize,
                    cleanup=cleanup,
                )(f),
            )
            attrs = MethodAttributes.for_method(target_wrapper, True, True)
            assert attrs is not None, "MethodAttributes were not written successfully."
            attrs.put(self.Attributes.TARGET)
            if default:
                attrs.put(self.Attributes.DEFAULT)
            if name is not None:
                target_wrapper = named(name)(target_wrapper)
            self.provide(target_wrapper)
            return target_wrapper

        if callable(name_or_f):
            return wrapper(name_or_f)

        return wrapper

    async def _resolve_tasks(self, config: Config) -> list[Recipe]:
        task_map: dict[str, Recipe] = await self.addressable_task_map()
        targets = config.targets

        if not targets:
            default_task = self.default_task()
            if default_task is not None:
                targets = [default_task]
            else:
                raise ValueError("No task specified and no default task defined.")

        task_names: set[str] = set()
        tasks = []
        for target in targets:
            results = fnmatch.filter(task_map.keys(), target)
            if len(results) < 1:
                raise ValueError(
                    f"Target filter `{target}` matched no defined targets."
                )

            for name in results:
                if name not in task_names:
                    task_names.add(name)
                    tasks.append(task_map[name])

        return tasks

    async def _print_help(self, config: Config, tasks: Iterable[Recipe]):
        parser = config._argparser()

        self.txt(f"# {self.name}")
        self.txt(parser.format_help(), render="dim")
        self.txt("")
        self.txt("# Target Tasks")
        self._list_tasks(config, tasks)

    def _list_tasks(self, config: Config, tasks: Iterable[Recipe]):
        tasks = sorted(tasks, key=lambda t: t.callsign)
        for t in tasks:
            self.txt.write(t.callsign, fg="cyan")
            if t.docs:
                self.txt.write(f" {t.fmt.symbols.target} ")
                self.txt.write(t.docs, render="dim")
            self.txt("")
        return tasks

    def _list_task_tree(
        self,
        tasks: Iterable[Recipe],
        indent=0,
        visited: Optional[set[Recipe]] = None,
    ):
        visited = visited or set()
        tasks = sorted((t for t in tasks if t not in visited), key=lambda t: t.callsign)
        new_visited = visited | set(tasks)

        first_indent = True
        for task in tasks:
            if indent > 0:
                self.txt.write("  " * indent)
                if first_indent:
                    self.txt.write(f" {self.TREE_SUB_ARROW} ", fg="white", render="dim")
                    first_indent = False
                else:
                    self.txt.write(f' {" " * len(self.TREE_SUB_ARROW)} ')

                self.txt.print(task.callsign)
            else:
                self.txt.print(task.callsign, fg="cyan", render="bold")

            if task not in visited:
                self._list_task_tree(task.children, indent + 1, new_visited)

        return 0

    async def build_async(self, config: Config) -> list[Any]:
        bus = EventBus.get()
        max_jobs = Shell.max_jobs

        try:
            Shell.set_max_jobs(config.jobs)

            try:
                mode_method = self._build_mode_methods[config.mode]

            except IndexError:
                raise RuntimeError("Unknown build mode encountered: {config.mode}")

            return await mode_method(config)

        finally:
            Shell.set_max_jobs(max_jobs)
            bus.shutdown()

    async def _build_mode_build(self, config: Config):
        tasks = await self._resolve_tasks(config)
        self.scan.scan_params(*tasks)

        try:
            while self.scan.has_recipes():
                await self.scan.gather_all()

            args = self.scan.args(Recipe.PassMode.RESULTS)
            self.scan.clear()
            self.txt("OK", fg="green", render="bold")
            return args

        except BuildError as e:
            self.txt.embrace("ERROR", fg="red", render="bold")
            self.txt("FAIL", fg="red", render="bold")
            raise e

    async def _build_mode_clean(self, config: Config):
        tasks = await self._resolve_tasks(config)
        match config.cleanup_mode:
            case Config.CleanupMode.SHALLOW:
                return await asyncio.gather(*[task.clean() for task in tasks])
            case Config.CleanupMode.RECURSIVE:
                return await asyncio.gather(
                    *[task.clean() for task in tasks],
                    *[task.clean_components(recursive=True) for task in tasks],
                )
            case _:
                raise ValueError("Config.cleanup_mode not specified.")

    async def _build_mode_help(self, config: Config):
        tasks = await self.tasks()
        parser = config._argparser()

        self.txt(f"# {self.name}")
        self.txt(parser.format_help(), render="dim")
        self.txt("")
        self.txt("# Target Tasks")
        self._list_tasks(config, tasks)

    async def _build_mode_list(self, config: Config):
        return self._list_tasks(config, await self.tasks())

    async def _build_mode_list_all(self, config: Config):
        return self._list_tasks(config, (await self.addressable_task_map()).values())

    async def _build_mode_query(self, config: Config):
        assert config.query is not None
        tasks = await self.tasks()
        task_names = set([t.callsign for t in tasks])
        query_names = config.query.split(",")
        missing_names = [name for name in query_names if name not in task_names]
        return len(missing_names)

    async def _build_mode_rebuild(self, config: Config):
        await self._build_mode_clean(config)
        return await self._build_mode_build(config)

    async def _build_mode_tree(self, config: Config):
        return self._list_task_tree(await self.tasks())

    async def _build_loop(self, bus, config: Config):
        result, _ = await asyncio.gather(self.build_async(config), bus.run())
        return result

    def build(self, *argv_, raise_errors=False):
        target_argv = [*argv_]
        build_argv = []

        while target_argv:
            arg = target_argv.pop(0)
            if arg == "@":
                break
            build_argv.append(arg)

        config = Config().parse_args(*build_argv)

        @self.provide
        def argv():
            return target_argv

        @self.provide
        def build_config():
            return config

        match config.color:
            case Config.ColorOptions.YES:
                enable_color()
            case Config.ColorOptions.NO:
                disable_color()
            case Config.ColorOptions.AUTO:
                pass

        try:
            with EventBus.session():
                bus = EventBus.get()
                for hook in self.bus_hooks:
                    hook(config, self, bus)
                return asyncio.run(self._build_loop(bus, config))

        except BuildError as e:
            if config.debug:
                traceback.print_exc()
            if raise_errors:
                raise e


# --------------------------------------------------------------------
class DefaultEngineHook:
    def __init__(self):
        self.spinner = Spinner("resolving")
        self.txt: Optional[TextDecorator] = None
        self.quiet = False
        self.to_stdout = False
        self.active_offset = -1
        self.active_inc_dt = datetime.min
        self.print_len = 0

    def embrace(self, *content, **kwargs):
        assert self.txt
        return self.txt.embrace(*content, **kwargs)

    def print(self, *content, **kwargs):
        assert self.txt
        return self.txt(*content, **kwargs)

    def sigil(self, event):
        return event.context.sigil().replace(
            Recipe.SIGIL_DELIMITER, event.context.fmt.symbols.target
        )

    async def on_frame(self, event):
        if self.quiet:
            return
        assert self.txt

        now = datetime.now()
        sorted_active = sorted(Recipe.active, key=lambda r: r.sigil())

        if now > self.active_inc_dt:
            self.active_offset += 1
            self.active_inc_dt = now + timedelta(seconds=1)

        if self.active_offset >= len(sorted_active):
            self.active_offset = 0

        if sorted_active:
            recipe = sorted_active[self.active_offset]
            message = f"resolving {len(Recipe.active)} [{recipe.sigil()}]"
        else:
            message = ""

        if message != self.spinner.message:
            self.txt.wipeline(self.print_len)
            self.spinner.message = message
        self.txt.wipe = self.print_len = await self.spinner.spin()

    def on_clean(self, event):
        if self.quiet:
            return
        self.embrace(self.sigil(event), fg="green", render="bold")
        self.print(event.context.fmt.clean(event.context))

    def on_error(self, event):
        if not self.quiet:
            self.embrace(self.sigil(event), fg="red", render="bold")
        self.print(event.data)

    def on_fail(self, event):
        if self.quiet:
            return
        self.embrace(self.sigil(event), fg="white", bg="red", render="bold")
        self.print(event.context.fmt.fail(event.context))

    def on_info(self, event: Event):
        if not self.quiet and not self.to_stdout:
            self.embrace(self.sigil(event), fg="white", render="dim")
        if self.to_stdout:
            print(event.data)
        else:
            self.print(event.data, fg="white", render="dim")

    def on_start(self, event: Event):
        if self.quiet:
            return
        self.embrace(self.sigil(event), fg="cyan", render="bold")
        self.print(event.context.fmt.start(event.context))

    def on_success(self, event: Event):
        if self.quiet:
            return
        self.embrace(self.sigil(event), fg="green", render="bold")
        self.print(event.context.fmt.ok(event.context))

    def on_warning(self, event: Event):
        if not self.quiet:
            self.embrace(self.sigil(event), fg="yellow", render="bold")
            self.print(event.data, fg="yellow", render="dim")
        else:
            self.print(event.data)

    def __call__(self, config: Config, engine: Engine, bus: EventBus):
        self.txt = engine.txt
        self.spinner.colorized = is_color_enabled()
        self.quiet = config.quiet
        self.to_stdout = config.to_stdout
        for event, listener in [
            (Events.CLEAN, self.on_clean),
            (Events.ERROR, self.on_error),
            (Events.FAIL, self.on_fail),
            (Events.INFO, self.on_info),
            (Events.START, self.on_start),
            (Events.SUCCESS, self.on_success),
            (Events.WARNING, self.on_warning),
            (EventBus.FRAME, self.on_frame),
        ]:
            bus.subscribe(event, listener)


# --------------------------------------------------------------------
engine = Engine()
engine.add_hook(DefaultEngineHook())
provide = engine.provide
recipe = engine.recipe
factory = engine.factory
task = engine.task


# --------------------------------------------------------------------
def build():
    result = engine.build(*sys.argv[1:])
    if isinstance(result, int):
        sys.exit(result)
    sys.exit(0)



# --------------------------------------------------------------------
# testing.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday April 13, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import io
import sys


# --------------------------------------------------------------------
class OutputCapture:
    VALID_TARGETS = set(["stdout", "stderr"])

    def __init__(self, **kwargs):
        self.targets = set()

        for k, v in kwargs.items():
            if k not in OutputCapture.VALID_TARGETS:
                raise ValueError(
                    f"Capture targets must be one or more of {OutputCapture.VALID_TARGETS}."
                )
            if v:
                self.targets.add(k)

        self.stdin = io.StringIO()
        self.stdout = io.StringIO()
        self.stderr = io.StringIO()
        self.original_streams: dict[str, io.IOBase] = {}

    def __enter__(self, *args):
        for target in self.targets:
            self.original_streams[target] = getattr(sys, target)
            setattr(sys, target, getattr(self, target))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for target in self.targets:
            setattr(sys, target, self.original_streams[target])
        self.original_streams.clear()



# --------------------------------------------------------------------
# shell.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Saturday October 24, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import itertools
import os
import shlex
import shutil
import signal
import subprocess
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Set, Tuple, Union

from xeno.utils import decode, is_iterable


# --------------------------------------------------------------------
class Environment(dict[str, str]):
    """
    An environment dictionary that knows how to append shell
    flag variables together when added to other dictionaries.
    """

    @classmethod
    def context(cls):
        return cls(os.environ)

    def select(self, *names: str, **defaults: Any) -> "Environment":
        filtered_env = Environment()
        for name in itertools.chain(names, defaults.keys()):
            if name in self:
                filtered_env[name] = self[name]
            elif name in defaults:
                filtered_env[name] = defaults[name]
            else:
                filtered_env[name] = ""
        return filtered_env

    def update(self, *args, **kwargs):
        for arg in args:
            if hasattr(arg, "keys"):
                self.update(**arg)
            else:
                for k, v in arg:
                    self.update(**{k: v})

        for k, v in kwargs.items():
            if is_iterable(v):
                self[k] = shlex.join([str(x) for x in v])
            else:
                self[k] = str(v)

        return self

    def append(self, *args, **kwargs):
        for arg in args:
            if hasattr(arg, "keys"):
                self.append(**arg)
            else:
                for k, v in arg:
                    self.append(**{k: v})

        for k, v in kwargs.items():
            if is_iterable(v):
                v = [str(x) for x in v]
            else:
                v = shlex.split(str(v))
            if k in self:
                v = shlex.split(self[k]) + v
            self[k] = shlex.join(v)
        return self

    def split(self, key: str, default: Optional[Sequence[str]] = None):
        if key in self:
            return shlex.split(self[key])
        else:
            return default or []

    def __setitem__(self, key: str, value: Any):
        if is_iterable(value):
            value = shlex.join([str(s) for s in value])
        super().__setitem__(key, str(value))

    def __add__(self, rhs: dict[str, str | Iterable[str]]) -> "Environment":
        new_env = Environment()
        for key, rhs_value in rhs.items():
            if key in self:
                lhs_value = self.split(key)
                if isinstance(rhs, Environment):
                    rhs_value = rhs.split(key)
                    new_env[key] = lhs_value + rhs_value
                elif is_iterable(rhs_value):
                    new_env[key] = lhs_value + [str(s) for s in rhs_value]
                else:
                    new_env[key] = lhs_value + shlex.split(rhs_value)
            else:
                new_env[key] = rhs_value
        return new_env

    def __iadd__(self, rhs: dict[str, str | Iterable[str]]) -> "Environment":
        new_dict = self + rhs
        super().update(new_dict)
        return self


# --------------------------------------------------------------------
EnvDict = Environment | Dict[str, Any]
InputSource = Callable[[], str]
LineSink = Callable[[str, asyncio.StreamWriter], None]
OutputTaskData = Tuple[asyncio.StreamReader, LineSink]
PathSpec = Union[str | Path]


# --------------------------------------------------------------------
def digest_env(env: EnvDict):
    flat_env: Dict[str, str] = {}
    for key, value in env.items():
        if is_iterable(value):
            value = " ".join(shlex.quote(str(s)) for s in value)
        flat_env[key] = str(value)
    return flat_env


# --------------------------------------------------------------------
def digest_params(params: EnvDict):
    flat_params: Dict[str, str] = {}
    for key, value in params.items():
        if is_iterable(value):
            value = " ".join(str(s) for s in value)
        flat_params[key] = str(value)
    return flat_params


# --------------------------------------------------------------------
def check(cmd: Union[str, Iterable[str]], **kwargs):
    if isinstance(cmd, str):
        args = shlex.split(cmd)
    else:
        args = [*cmd]
    return subprocess.check_output(args, **kwargs).decode("utf-8").strip()


# --------------------------------------------------------------------
def remove_paths(*paths: Path, as_user: Optional[str] = None):
    for path in paths:
        if as_user is not None:
            result = Shell().interact_as(as_user, ["rm", "-rf", str(path.absolute())])
            if result != 0:
                raise RuntimeError(f"Failed to remove path `f{path}` as `f{as_user}`.")

        else:
            if not path.exists():
                continue

            if path.is_dir() and not path.is_symlink():
                shutil.rmtree(path)
            else:
                path.unlink()


# --------------------------------------------------------------------
class Shell:
    max_jobs: int = 0
    job_semaphore: asyncio.Semaphore = asyncio.Semaphore(max_jobs)

    @classmethod
    def set_max_jobs(cls, n: int):
        cls.max_jobs = n
        cls.job_semaphore = asyncio.Semaphore(cls.max_jobs)

    def __init__(self, env: EnvDict = dict(os.environ), cwd: Optional[PathSpec] = None):
        self._env = digest_env(env)
        self._cwd = Path(cwd) if cwd is not None else Path.cwd()

    def env(self, new_env: EnvDict):
        return Shell({**self._env, **new_env}, self._cwd)

    def cd(self, new_cwd: Path):
        assert new_cwd.exists() and new_cwd.is_dir(), "Invalid directory provided."
        return Shell(self._env, new_cwd)

    # pylint: disable=no-member
    # see: https://github.com/PyCQA/pylint/issues/1469
    async def _create_proc(self, cmd: str) -> asyncio.subprocess.Process:
        return await asyncio.create_subprocess_shell(
            cmd,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=self._env,
            cwd=self._cwd,
            shell=True,
        )

    def _interact(self, cmd: str, check: bool, ctrlc: bool) -> int:
        existing_handler = signal.getsignal(signal.SIGINT)
        ctrlc_happened = [False]

        def ctrlc_handler(*_):
            print()
            ctrlc_happened[0] = True

        try:
            if ctrlc:
                signal.signal(signal.SIGINT, ctrlc_handler)

            returncode = subprocess.call(cmd, env=self._env, cwd=self._cwd, shell=True)
            if ctrlc_happened[0]:
                return 0
            assert not check or returncode == 0, "Command failed."
            return returncode

        finally:
            if ctrlc:
                signal.signal(signal.SIGINT, existing_handler)

    def interpolate(
        self,
        cmd: Union[str, Iterable[str]],
        params: EnvDict,
        wrappers: Dict[str, Callable[[str], str]] = {},
        redacted: Set[str] = set(),
    ) -> str:
        digested_params = {
            k: wrappers[k](v)
            if k in wrappers
            else (wrappers["*"](v) if "*" in wrappers else v)
            for k, v in digest_params(params).items()
        }

        redacted_params = {
            k: v if k not in redacted else "<redacted>"
            for k, v in digested_params.items()
        }

        if isinstance(cmd, str):
            final_cmd = cmd.format(**self._env, **redacted_params)
        else:
            final_cmd = shlex.join(
                [str(c).format(**self._env, **redacted_params) for c in cmd]
            )

        return final_cmd

    async def _run(
        self,
        cmd: Union[str, Iterable[str]],
        stdin: Optional[InputSource] = None,
        stdout: Optional[LineSink] = None,
        stderr: Optional[LineSink] = None,
        check=False,
        **params,
    ) -> int:
        readline_tasks: Dict[asyncio.Future[Any], OutputTaskData] = {}

        def setup_readline_task(stream: asyncio.StreamReader, sink: LineSink):
            readline_tasks[asyncio.Task(stream.readline())] = (stream, sink)

        cmd = self.interpolate(cmd, params)
        proc = await self._create_proc(cmd)
        assert proc.stdout is not None
        assert proc.stderr is not None
        if stdin:
            assert proc.stdin
            proc.stdin.write(stdin().encode("utf-8"))
        if stdout:
            setup_readline_task(proc.stdout, stdout)
        if stderr:
            setup_readline_task(proc.stderr, stderr)

        while readline_tasks:
            done, _ = await asyncio.wait(
                readline_tasks, return_when=asyncio.FIRST_COMPLETED
            )

            for future in done:
                stream, sink = readline_tasks.pop(future)
                line = future.result()
                if line:
                    line = decode(line).rstrip()
                    assert proc.stdin is not None
                    sink(line, proc.stdin)
                    setup_readline_task(stream, sink)

        await proc.wait()
        assert proc.returncode is not None, "proc.returncode is None"
        assert not check or proc.returncode == 0, "Command failed."
        return proc.returncode

    async def run(
        self,
        cmd: Union[str, Iterable[str]],
        stdin: Optional[InputSource] = None,
        stdout: Optional[LineSink] = None,
        stderr: Optional[LineSink] = None,
        check=False,
        **params,
    ) -> int:
        if self.max_jobs < 1:
            return await self._run(cmd, stdin, stdout, stderr, check, **params)
        else:
            async with self.job_semaphore:
                return await self._run(cmd, stdin, stdout, stderr, check, **params)

    def sync(
        self,
        cmd: Union[str, Iterable[str]],
        stdin: Optional[InputSource] = None,
        stdout: Optional[LineSink] = None,
        stderr: Optional[LineSink] = None,
        check=False,
        **params,
    ) -> int:
        return asyncio.run(self.run(cmd, stdin, stdout, stderr, check, **params))

    def interact(
        self, cmd: Union[str, Iterable[str]], check=False, ctrlc=False, **params
    ) -> int:
        cmd = self.interpolate(cmd, params)
        return self._interact(cmd, check, ctrlc)

    def interact_as(
        self,
        as_user: str,
        cmd: Union[str, Iterable[str]],
        check=False,
        ctrlc=False,
        **params,
    ) -> int:
        cmd = self.interpolate(cmd, params)
        cmd = shlex.join(["sudo", "-u", as_user, "sh", "-c", cmd])
        return self._interact(cmd, check, ctrlc)


# --------------------------------------------------------------------
def select_env(*args, **kwargs):
    return Environment.context().select(*args, **kwargs)


# --------------------------------------------------------------------
get_env = Environment.context



# --------------------------------------------------------------------
# attributes.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import inspect
from typing import Any, List, cast

from .errors import InjectionError
from .utils import bind_unbound_method, get_params_from_signature

# --------------------------------------------------------------------
NOTHING = object()

# --------------------------------------------------------------------
class Tags:
    ALIASES = "xeno.tags.aliases"
    CONST_MAP = "xeno.tags.const_map"
    DOCS = "xeno.tags.docs"
    INJECTION_POINT = "xeno.tags.injection_point"
    NAME = "xeno.tags.name"
    NAMESPACE = "xeno.tags.namespace"
    PARAMS = "xeno.tags.params"
    PROVIDER = "xeno.tags.provider"
    QUALNAME = "xeno.tags.qualname"
    RESOURCE_FULL_NAME = "xeno.tags.resource_full_name"
    SINGLETON = "xeno.tags.singleton"
    USING_NAMESPACES = "xeno.tags.using_namespaces"


# --------------------------------------------------------------------
class Attributes:
    def __init__(self):
        self.attr_map = {}

    @staticmethod
    def for_object(obj, create=True, write=False, factory=lambda x: Attributes()):
        try:
            return obj._attrs

        except AttributeError:
            if create:
                attrs = factory(obj)
                if write:
                    obj._attrs = attrs
                return attrs
            return None

    def put(self, attr, value: Any = True):
        self.attr_map[attr] = value
        return self

    def get(self, attr, default_value=NOTHING):
        if attr in self.attr_map:
            return self.attr_map[attr]
        if default_value is NOTHING:
            raise AttributeError("No such attribute: %s" % attr)
        return default_value

    def check(self, attr):
        return bool(self.get(attr, None))

    def has(self, attr):
        return self.get(attr, None) is not None

    def merge(self, attr):
        self.attr_map.update(attr.attr_map)
        return self


# --------------------------------------------------------------------
class ClassAttributes(Attributes):
    @staticmethod
    def for_class(class_, create=True, write=False):
        return Attributes.for_object(
            class_, create, write, factory=lambda x: ClassAttributes(x)
        )

    @staticmethod
    def for_object(obj, create=True, write=False):
        return ClassAttributes.for_class(obj.__class__, create, write)

    def __init__(self, class_):
        super().__init__()
        self.put(Tags.NAME, class_.__name__)
        self.put(Tags.QUALNAME, class_.__qualname__)
        self.put(Tags.DOCS, class_.__doc__)


# --------------------------------------------------------------------
class MethodAttributes(Attributes):
    @staticmethod
    def for_method(f, create=True, write=False):
        return Attributes.for_object(
            f, create, write, factory=lambda x: MethodAttributes(x)
        )

    @staticmethod
    def wraps(f1):
        def decorator(f2):
            attr1 = MethodAttributes.for_method(f1)
            attr2 = MethodAttributes.for_method(f2, write=True)
            assert attr2 is not None
            if attr1 is not None:
                attr2.merge(attr1)
            return f2

        return decorator

    @staticmethod
    def add(name, value=True):
        def decorator(f):
            attrs = MethodAttributes.for_method(f, write=True)
            assert attrs is not None
            attrs.put(name, value)
            return f

        return decorator

    def __init__(self, f):
        super().__init__()
        self.put(Tags.NAME, f.__name__)
        self.put(Tags.QUALNAME, f.__qualname__)
        self.put(Tags.PARAMS, get_params_from_signature(f))
        self.put(Tags.DOCS, f.__doc__)


# --------------------------------------------------------------------
def get_injection_params(f, unbound_ctor=False):
    """
    Fetches the injectable parameter names of parameters to the given
    method, along with a set of parameters which have default values
    and should be considered optional dependencies.

    This method will throw InjectionError if the method provided has
    arguments that are not POSITIONAL_OR_KEYWORD or KEYWORD_ONLY.

    If the method provided is an unbound object constructor,
    unbound_ctor must be set to True to prevent 'self' from being
    returned by this method as an injectable parameter.
    """
    injection_param_names = []
    default_param_set = set()
    params: List = []

    if not inspect.ismethod(f) and unbound_ctor and not inspect.isfunction(f):
        # We do not want to try to inject a slot wrapper
        # version of __init__, as its params are (*args, **kwargs)
        # and it does nothing anyway.
        return [], set()

    attr = MethodAttributes.for_method(f)
    assert attr is not None
    if attr.has(Tags.PARAMS):
        params = cast(list, attr.get(Tags.PARAMS))
    else:
        params = get_params_from_signature(f)

    if inspect.ismethod(f) or unbound_ctor:
        # Don't try to inject the 'self' parameter of an
        # unbound constructor.
        params = params[1:]

    for param in params:
        if param.kind in [
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ]:
            if param.default != param.empty:
                default_param_set.add(param.name)
            injection_param_names.append(param.name)
        else:
            raise InjectionError(
                "Xeno only supports injection of POSITIONAL_OR_KEYWORD and "
                "KEYWORD_ONLY arguments, %s arguments (%s of %s) "
                "are not supported." % (param.kind, param.name, f.__qualname__)
            )
    return injection_param_names, default_param_set


# --------------------------------------------------------------------
def scan_methods(obj, filter_f):
    """
    Scan the object for methods that match the given attribute filter
    and return them as a stream of tuples.
    """
    for class_ in inspect.getmro(obj.__class__):
        for _, method in inspect.getmembers(class_, predicate=inspect.isfunction):
            attrs = MethodAttributes.for_method(method, create=False)
            if attrs is not None and filter_f(attrs):
                yield (attrs, bind_unbound_method(obj, method))


# --------------------------------------------------------------------
def get_injection_points(obj):
    """
    Scan the object and all of its parents for injection points
    and return them as a stream of tuples.
    """

    return scan_methods(obj, lambda attr: attr.check(Tags.INJECTION_POINT))


# --------------------------------------------------------------------
def get_providers(obj):
    """
    Scan the object and all of its parents for providers and return
    them as a stream of tuples.
    """

    return scan_methods(obj, lambda attr: attr.check(Tags.PROVIDER))



from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="xeno",
    version="7.4.4",
    description="The Python IOC app and build framework from outer space.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lainproliant/xeno",
    author="Lain Musgrove (lainproliant)",
    author_email="lainproliant@gmail.com",
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="IOC dependency injector build system",
    packages=find_packages(),
    install_requires=[],
    extras_require={},
    package_data={"xeno": ["LICENSE"]},
    data_files=[],
    entry_points={"console_scripts": []},
)



# -------------------------------------------------------------------
# Xeno: The Python dependency injector from outer space.
#
# Author: Lain Musgrove (lainproliant)
# Date: Sunday, August 28th 2016
#
# Released under a 3-clause BSD license, see LICENSE for more info.
# -------------------------------------------------------------------

__all__ = [
    "InjectionError",
    "MissingResourceError",
    "MissingDependencyError",
    "MethodInjectionError",
    "ClassInjectionError",
    "CircularDependencyError",
    "UndefinedNameError",
    "UnknownNamespaceError",
    "InvalidResourceError",
    "ClassAttributes",
    "MethodAttributes",
    "Namespace",
    "Tags",
    "async_map",
    "async_wrap",
    "singleton",
    "provide",
    "inject",
    "named",
    "alias",
    "namespace",
    "const",
    "using",
    "scan_methods",
    "get_injection_points",
    "get_providers",
    "get_injection_params",
    "Injector",
    "AsyncInjector",
    "SyncInjector",
    "NOTHING",
]

from .async_injector import AsyncInjector
from .attributes import (
    NOTHING,
    ClassAttributes,
    MethodAttributes,
    get_injection_params,
    get_injection_points,
    get_providers,
    scan_methods,
)
from .decorators import (
    Tags,
    alias,
    const,
    inject,
    named,
    namespace,
    provide,
    singleton,
    using,
)
from .errors import (
    CircularDependencyError,
    ClassInjectionError,
    InjectionError,
    InvalidResourceError,
    MethodInjectionError,
    MissingDependencyError,
    MissingResourceError,
    UndefinedNameError,
    UnknownNamespaceError,
)
from .namespaces import Namespace
from .sync_injector import SyncInjector
from .utils import async_map, async_wrap

# For backwards compatibility with older versions of Xeno.
Injector = AsyncInjector



# --------------------------------------------------------------------
# namespace.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday May 7, 2020
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------


# --------------------------------------------------------------------
class Namespace:
    ROOT = "@root"
    SEP = "/"

    @staticmethod
    def join(*args):
        return Namespace.SEP.join(args)

    @staticmethod
    def root():
        return Namespace(Namespace.ROOT)

    @staticmethod
    def leaf_name(name):
        return name.split(Namespace.SEP)[-1]

    def __init__(self, name):
        self.name = name
        self.sub_namespaces = {}
        self.leaves = set()

    def add(self, name):
        if not name:
            raise ValueError("Leaf node name is empty!")
        parts = name.split(Namespace.SEP)
        if len(parts) == 1:
            if name in self.sub_namespaces:
                raise ValueError(
                    'Leaf node cannot have the same name as an existing '
                    'namespace: "%s"'
                    % name
                )
            self.leaves.add(name)
        else:
            if not parts[0] in self.sub_namespaces:
                if parts[0] in self.leaves:
                    raise ValueError(
                        'Namespace cannot have the same name as an existing '
                        'leaf node: "%s"'
                        % parts[0]
                    )
                self.sub_namespaces[parts[0]] = Namespace(parts[0])
            namespace = self.sub_namespaces[parts[0]]
            namespace.add(Namespace.join(*parts[1:]))

    def add_namespace(self, name):
        if not name:
            raise ValueError("Namespace name is empty!")
        parts = name.split(Namespace.SEP)
        ns = self
        for part in parts:
            if part in ns.sub_namespaces:
                ns = ns.sub_namespaces[part]
            else:
                new_ns = Namespace(part)
                ns.sub_namespaces[part] = new_ns
                ns = new_ns

    def get_namespace(self, name=None):
        if name == Namespace.SEP or not name:
            return self
        if name.startswith(Namespace.SEP):
            return self.get_namespace(name[1:])
        nodes = name.split(Namespace.SEP)
        if nodes[0] in self.sub_namespaces:
            return self.sub_namespaces[nodes[0]].get_namespace(
                Namespace.SEP.join(nodes[1:])
            )
        return None

    def get_leaves(self, recursive=False, prefix=""):
        if not recursive:
            return list(self.leaves)
        if self.name == Namespace.ROOT:
            prefix = ""
        else:
            prefix += self.name + Namespace.SEP

        leaves = []
        leaves.extend([prefix + x for x in self.leaves])
        for ns in self.sub_namespaces.values():
            leaves.extend(ns.get_leaves(True, prefix))
        return leaves



# --------------------------------------------------------------------
# spinner.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday January 12, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import itertools
import sys
from datetime import datetime, timedelta
from typing import Iterable, Optional

from xeno.color import color, is_enabled as is_color_enabled

DEFAULT_SHAPE = [
    "[=   ]",
    "[==  ]",
    "[=== ]",
    "[====]",
    "[ ===]",
    "[  ==]",
    "[   =]",
    "[    ]",
]


class Spinner:
    def __init__(self, message: str, interval: float = 0.05, delay: float = 0.25):
        self.message = message
        self.colorized = is_color_enabled
        self.cycle: Optional[Iterable[str]] = None
        self.interval = interval
        self.delay = delay
        self.start = datetime.now()

    def _setup_cycle(self):
        shape_array = []

        if self.colorized:
            for shape in DEFAULT_SHAPE:
                shape_array.append(
                    "".join(
                        [
                            shape[0],
                            color(shape[1:-1], fg="yellow"),
                            shape[-1],
                        ]
                    )
                )
        else:
            for shape in DEFAULT_SHAPE:
                shape_array.extend(DEFAULT_SHAPE)

        self.cycle = itertools.cycle(shape_array)

    async def spin(self) -> int:
        if not sys.stdout.isatty():
            return 0

        if not self.cycle:
            self._setup_cycle()

        assert self.cycle is not None

        erase_chars = 0
        if datetime.now() - self.start > timedelta(seconds=self.delay):
            sys.stdout.write("\r")
            erase_chars = sys.stdout.write(
                next(self.cycle) + color(f" {self.message} ", render="dim")
            )
            sys.stdout.flush()
        await asyncio.sleep(self.interval)
        return erase_chars
